# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Any, Callable, List, Optional, Type, Union

import numpy as np
import numpy.typing as npt

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
    ProcessorPluginPreset,
    ProcessorViewPluginBase,
    SetupMessage,
)
from acconeer.exptool.a121.algo.touchless_button import (
    MeasurementType,
    Processor,
    ProcessorConfig,
    ProcessorResult,
    RangeResult,
    get_close_and_far_processor_config,
    get_close_and_far_sensor_config,
    get_close_processor_config,
    get_close_sensor_config,
    get_far_processor_config,
    get_far_sensor_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    pidgets,
)
from acconeer.exptool.app.new.ui.components.pidgets.hooks import (
    disable_if,
    parameter_is,
)


log = logging.getLogger(__name__)


class PluginPresetId(Enum):
    CLOSE_RANGE = auto()
    FAR_RANGE = auto()
    CLOSE_AND_FAR_RANGE = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):
    PLUGIN_PRESETS = {
        PluginPresetId.CLOSE_RANGE.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_close_sensor_config()),
            processor_config=get_close_processor_config(),
        ),
        PluginPresetId.FAR_RANGE.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_far_sensor_config()),
            processor_config=get_far_processor_config(),
        ),
        PluginPresetId.CLOSE_AND_FAR_RANGE.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_close_and_far_sensor_config()),
            processor_config=get_close_and_far_processor_config(),
        ),
    }

    @classmethod
    def get_processor(cls, state: ProcessorBackendPluginSharedState[ProcessorConfig]) -> Processor:
        if state.metadata is None:
            msg = "metadata is None"
            raise RuntimeError(msg)

        if isinstance(state.metadata, list):
            msg = "metadata is unexpectedly extended"
            raise RuntimeError(msg)

        return Processor(
            sensor_config=state.session_config.sensor_config,
            processor_config=state.processor_config,
            metadata=state.metadata,
        )

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_close_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "sensitivity_close": pidgets.FloatPidgetFactory(
                name_label_text="Sensitivity (close range):",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "sensitivity_close"),
                decimals=1,
                limits=(0.1, 4),
                hooks=disable_if(parameter_is("measurement_type", MeasurementType.FAR_RANGE)),
            ),
            "patience_close": pidgets.IntPidgetFactory(
                name_label_text="Patience (close range):",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "patience_close"),
                limits=(0, None),
                hooks=disable_if(parameter_is("measurement_type", MeasurementType.FAR_RANGE)),
            ),
            "sensitivity_far": pidgets.FloatPidgetFactory(
                name_label_text="Sensitivity (far range):",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "sensitivity_far"),
                decimals=1,
                limits=(0.1, 4),
                hooks=disable_if(parameter_is("measurement_type", MeasurementType.CLOSE_RANGE)),
            ),
            "patience_far": pidgets.IntPidgetFactory(
                name_label_text="Patience (far range):",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "patience_far"),
                limits=(0, None),
                hooks=disable_if(parameter_is("measurement_type", MeasurementType.CLOSE_RANGE)),
            ),
            "calibration_duration_s": pidgets.FloatPidgetFactory(
                name_label_text="Calibration duration:",
                name_label_tooltip=get_attribute_docstring(
                    ProcessorConfig, "calibration_duration_s"
                ),
                suffix="s",
                limits=(0, None),
            ),
            "calibration_interval_s": pidgets.FloatPidgetFactory(
                name_label_text="Calibration interval:",
                name_label_tooltip=get_attribute_docstring(
                    ProcessorConfig, "calibration_interval_s"
                ),
                suffix="s",
                limits=(1, None),
            ),
            "measurement_type": pidgets.EnumPidgetFactory(
                enum_type=MeasurementType,
                name_label_text="Range:",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "measurement_type"),
                label_mapping={
                    MeasurementType.CLOSE_RANGE: "Close",
                    MeasurementType.FAR_RANGE: "Far",
                    MeasurementType.CLOSE_AND_FAR_RANGE: "Close and far",
                },
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def supports_multiple_subsweeps(self) -> bool:
        return True


class PlotPlugin(PgPlotPlugin):
    score_history_close: Optional[npt.NDArray[np.float64]]
    score_history_far: Optional[npt.NDArray[np.float64]]
    score_history_curves_close: Optional[npt.NDArray[np.object_]]
    score_history_curves_far: Optional[npt.NDArray[np.object_]]
    score_history_curves: Union[List[npt.NDArray[Any]], npt.NDArray[Any]]

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[ProcessorResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            if isinstance(message.metadata, list):
                msg = "Metadata is unexpectedly extended"
                raise RuntimeError(msg)

            self.setup(
                metadata=message.metadata,
                sensor_config=message.session_config.sensor_config,
                processor_config=message.processor_config,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(processor_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        metadata: a121.Metadata,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
    ) -> None:
        self.plot_layout.clear()

        self.detection_history_plot = self._create_detection_plot(self.plot_layout)
        detection_plot_legend = self.detection_history_plot.legend

        self.detection_history_curve_close = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(1, width=5)
        )
        self.detection_history_curve_far = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=5)
        )

        close_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Close detection")
        )
        far_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Far detection")
        )
        self.close_text_item = pg.TextItem(
            html=close_html,
            fill=pg.mkColor(0xFF, 0x7F, 0x0E),
            anchor=(0.5, 0),
        )
        self.far_text_item = pg.TextItem(
            html=far_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4),
            anchor=(0.5, 0),
        )
        pos_left = (100 / 3, 1.8)
        pos_right = (2 * 100 / 3, 1.8)
        self.close_text_item.setPos(*pos_left)
        self.far_text_item.setPos(*pos_right)
        self.detection_history_plot.addItem(self.close_text_item)
        self.detection_history_plot.addItem(self.far_text_item)
        self.close_text_item.hide()
        self.far_text_item.hide()

        self.detection_history = np.full((2, 100), np.nan)

        self.score_history_plot = self._create_score_plot(self.plot_layout)
        score_plot_legend = self.score_history_plot.legend
        self.score_smooth_max = et.utils.SmoothMax()

        self.threshold_history_curve_close = self.score_history_plot.plot(
            pen=et.utils.pg_pen_cycler(1, width=2.5, style="--"),
        )
        self.threshold_history_curve_far = self.score_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=2.5, style="--"),
        )

        self.threshold_history = np.full((2, 100), np.nan)

        cycle_index = 2  # To not have same colors as thresholds
        if processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            measurement_type = "Close"
            score_plot_legend.addItem(self.threshold_history_curve_close, "Close range threshold")
            detection_plot_legend.addItem(self.detection_history_curve_close, "Close range")
        elif processor_config.measurement_type == MeasurementType.FAR_RANGE:
            measurement_type = "Far"
            score_plot_legend.addItem(self.threshold_history_curve_far, "Far range threshold")
            detection_plot_legend.addItem(self.detection_history_curve_far, "Far range")

        if processor_config.measurement_type != MeasurementType.CLOSE_AND_FAR_RANGE:
            score_history = np.full((sensor_config.subsweep.num_points, 100), np.nan)
            score_history_curves = np.empty((sensor_config.subsweep.num_points,), dtype=object)
            for i in range(sensor_config.subsweep.num_points):
                score_history_curves[i] = pg.ScatterPlotItem(
                    brush=et.utils.pg_brush_cycler(cycle_index),
                    name=f"{measurement_type} range, point {i}",
                )
                self.score_history_plot.addItem(score_history_curves[i])
                cycle_index += 1

            if processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
                self.score_history_close = score_history
                self.score_history_curves_close = score_history_curves
                self.score_history_far = None
                self.score_history_curves_far = None
            elif processor_config.measurement_type == MeasurementType.FAR_RANGE:
                self.score_history_close = None
                self.score_history_curves_close = None
                self.score_history_far = score_history
                self.score_history_curves_far = score_history_curves

        elif processor_config.measurement_type == MeasurementType.CLOSE_AND_FAR_RANGE:
            score_plot_legend.addItem(self.threshold_history_curve_close, "Close range threshold")
            score_plot_legend.addItem(self.threshold_history_curve_far, "Far range threshold")
            detection_plot_legend.addItem(self.detection_history_curve_close, "Close range")
            detection_plot_legend.addItem(self.detection_history_curve_far, "Far range")
            self.score_history_close = np.full(
                (sensor_config.subsweeps[0].num_points, 100), np.nan
            )
            self.score_history_far = np.full((sensor_config.subsweeps[1].num_points, 100), np.nan)
            self.score_history_curves_close = np.empty(
                (sensor_config.subsweeps[0].num_points,), dtype=object
            )
            self.score_history_curves_far = np.empty(
                (sensor_config.subsweeps[1].num_points,), dtype=object
            )

            range_labels = ["Close", "Far"]
            for n, subsweep in enumerate(sensor_config.subsweeps):
                measurement_type = range_labels[n]
                score_history_curve_list = [
                    self.score_history_curves_close,
                    self.score_history_curves_far,
                ]
                for i in range(subsweep.num_points):
                    score_history_curve_list[n][i] = pg.ScatterPlotItem(
                        brush=et.utils.pg_brush_cycler(cycle_index),
                        name=f"{measurement_type} range, point {i}",
                    )
                    self.score_history_plot.addItem(score_history_curve_list[n][i])
                    cycle_index += 1
            self.score_history_curves_close = score_history_curve_list[0]
            self.score_history_curves_far = score_history_curve_list[1]

    def draw_plot_job(self, processor_result: ProcessorResult) -> None:
        def is_none_or_detection(x: Optional[RangeResult]) -> Optional[bool]:
            return x.detection if x is not None else None

        detection = np.array(
            [
                is_none_or_detection(processor_result.close),
                is_none_or_detection(processor_result.far),
            ]
        )
        self.detection_history = np.roll(self.detection_history, -1, axis=1)
        self.detection_history[:, -1] = detection

        self.detection_history_curve_close.setData(self.detection_history[0])
        self.detection_history_curve_far.setData(self.detection_history[1])

        if processor_result.close is not None:
            if processor_result.close.detection:
                self.close_text_item.show()
            else:
                self.close_text_item.hide()

        if processor_result.far is not None:
            if processor_result.far.detection:
                self.far_text_item.show()
            else:
                self.far_text_item.hide()

        max_val = 0.0

        def is_none_or_threshold(x: Optional[RangeResult]) -> Optional[float]:
            return x.threshold if x is not None else None

        threshold = np.array(
            [
                is_none_or_threshold(processor_result.close),
                is_none_or_threshold(processor_result.far),
            ]
        )
        if np.nanmax(np.array(threshold, dtype=float)) > max_val:
            max_val = np.nanmax(np.array(threshold, dtype=float))
        self.threshold_history = np.roll(self.threshold_history, -1, axis=1)
        self.threshold_history[:, -1] = threshold

        self.threshold_history_curve_close.setData(self.threshold_history[0])
        self.threshold_history_curve_far.setData(self.threshold_history[1])

        if self.score_history_close is not None:
            self.score_history_close = np.roll(self.score_history_close, -1, axis=1)
            assert processor_result.close is not None
            # Plot the second highest score
            self.score_history_close[:, -1] = np.sort(processor_result.close.score, axis=0)[-2, :]

            assert self.score_history_curves_close is not None
            for i, curve in enumerate(self.score_history_curves_close):
                # Assign x-values so that setData() doesn't give error when y-values are NaN
                curve.setData(np.arange(0, 100), self.score_history_close[i, :].flatten())

            if np.max(processor_result.close.score) > max_val:
                max_val = float(np.max(processor_result.close.score))

        if self.score_history_far is not None:
            self.score_history_far = np.roll(self.score_history_far, -1, axis=1)
            assert processor_result.far is not None
            # Plot the second highest score
            self.score_history_far[:, -1] = np.sort(processor_result.far.score, axis=0)[-2, :]

            assert self.score_history_curves_far is not None
            for i, curve in enumerate(self.score_history_curves_far):
                # Assign x-values so that setData() doesn't give error when y-values are NaN
                curve.setData(np.arange(0, 100), self.score_history_far[i, :].flatten())

            if np.max(processor_result.far.score) > max_val:
                max_val = float(np.max(processor_result.far.score))

        if max_val != 0.0:
            self.score_history_plot.setYRange(0.0, self.score_smooth_max.update(max_val))

    @staticmethod
    def _create_detection_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        detection_history_plot = parent.addPlot(row=0, col=0)
        detection_history_plot.setTitle("Detection")
        detection_history_plot.setLabel(axis="bottom", text="Frames")
        detection_history_plot.addLegend()
        detection_history_plot.setMenuEnabled(False)
        detection_history_plot.setMouseEnabled(x=False, y=False)
        detection_history_plot.hideButtons()
        detection_history_plot.showGrid(x=True, y=True, alpha=0.5)
        detection_history_plot.setYRange(-0.1, 1.8)
        return detection_history_plot

    @staticmethod
    def _create_score_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        score_history_plot = parent.addPlot(row=1, col=0)
        score_history_plot.setTitle("Detection score")
        score_history_plot.setLabel(axis="bottom", text="Frames")
        score_history_plot.addLegend()
        score_history_plot.setMenuEnabled(False)
        score_history_plot.setMouseEnabled(x=False, y=False)
        score_history_plot.hideButtons()
        score_history_plot.showGrid(x=True, y=True, alpha=0.5)
        return score_history_plot


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


TOUCHLESS_BUTTON_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="touchless_button",
    title="Touchless button",
    docs_link="https://docs.acconeer.com/en/latest/ref_apps/a121/touchless_button.html",
    description="Detect tap/wave motion and register as button press.",
    family=PluginFamily.REF_APP,
    presets=[
        PluginPresetBase(
            name="Close range",
            description="Close range",
            preset_id=PluginPresetId.CLOSE_RANGE,
        ),
        PluginPresetBase(
            name="Far range",
            description="Far range",
            preset_id=PluginPresetId.FAR_RANGE,
        ),
        PluginPresetBase(
            name="Close and far range",
            description="Close and far range",
            preset_id=PluginPresetId.CLOSE_AND_FAR_RANGE,
        ),
    ],
    default_preset_id=PluginPresetId.CLOSE_RANGE,
)



# Copyright (c) Acconeer AB, 2022-2026
# All rights reserved

"""Tool dfu/flash devices with bootloader firmware"""

import abc
import logging
import time

import serial
from serial.serialutil import SerialException
from serial.tools import list_ports

from acconeer.exptool._core.communication.comm_devices import SerialDevice
from acconeer.exptool._core.communication.links.usb_link import PyUsbCdc
from acconeer.exptool.flash._device_flasher_base import DeviceFlasherBase
from acconeer.exptool.flash._flash_exception import FlashException

from ._bootloader_comm import BLCommunication, CommandFailed


log = logging.getLogger(__name__)


class ImageFlasher:
    """A class that handles image flashing"""

    def __init__(self, serial_dev):
        self._ser = serial_dev
        self._comm = BLCommunication(self._ser)
        self._progress_callback = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset_and_close()

    def reset_and_close(self):
        """reset system and close communication channel"""
        if self._comm is not None:
            self._comm.reset()

        if self._ser is not None:
            self._ser.close()
            self._ser = None

    def set_progress_callback(self, progress_callback):
        self._progress_callback = progress_callback

    def erase_image(self, erase_size):
        """erase image"""
        if not self._comm.is_image_erased(erase_size):
            log.debug("-Flash not erased, erasing...")
            self._comm.image_erase(erase_size)
        else:
            log.debug("-Flash already erased")

    def flash_image_bytearray(self, image_data):
        """flash image byte array to bootloader board"""

        image_size = len(image_data)
        self.erase_image(image_size)

        max_chunk_size = 32768
        offset = 0
        while offset < image_size:
            chunk_size = min(image_size - offset, max_chunk_size)
            chunk_array = image_data[offset : offset + chunk_size]
            log.debug(
                f"\r-Flashing: Writing {chunk_size} bytes to offset {offset}                ",
            )
            if self._progress_callback is not None:
                self._progress_callback(int(100 * offset / image_size))
            self._comm.image_write_block(offset, chunk_array)
            offset += chunk_size
        if self._progress_callback is not None:
            self._progress_callback(100, True)

        log.debug("Flashed Image:")
        log.debug(f" - Name:    {self._comm.get_app_sw_name()}")
        log.debug(f" - Version: {self._comm.get_app_sw_version()}")

    def flash_image_file(self, image_name):
        """flash image file to bootloader board"""

        with open(image_name, "rb") as image_file:
            image_data = image_file.read()
            self.flash_image_bytearray(image_data)


class BootloaderTool(DeviceFlasherBase):
    """Class to handle dfu/flash of devices with bootloader firmware"""

    """
    Class to handle dfu/flash of devices with bootloader firmware

    Args:
    device_vid          - The USB vendor ID of the product(s)
    bootloader_pid      - The USB product ID of the product bootloader
    """

    def __init__(self, *, device_vid, bootloader_pid, is_usb=True):
        super().__init__()
        self.device_vid = device_vid
        self.bootloader_pid = bootloader_pid
        self.is_usb = is_usb

    @staticmethod
    def _find_port(search_vid, search_pid, search_port=None):
        all_ports = list_ports.comports()
        found_ports = [
            port for port in all_ports if port.vid == search_vid and port.pid == search_pid
        ]

        if len(found_ports) > 1 and search_port is not None:
            found_ports = [port for port in all_ports if port.device == search_port]
            if len(found_ports) < 1:
                msg = "Couldn't find a device on the specified port"
                raise FlashException(msg)
            return found_ports[0].device
        elif len(found_ports) > 0:
            return found_ports[0].device

        return None

    @staticmethod
    def _get_pid_from_port(port_path):
        all_ports = list_ports.comports()
        found_ports = [port for port in all_ports if port.device == port_path]

        if len(found_ports) != 1:
            msg = "Couldn't find a device on the specified port"
            raise FlashException(msg)
        return found_ports[0].pid

    @staticmethod
    def exploration_server_to_dfu(vid, pid):
        ser = PyUsbCdc(
            vid=vid,
            pid=pid,
        )
        ser.send_break()
        time.sleep(0.1)
        ser.write(bytes('{ "cmd": "stop_application" }\n', "utf-8"))
        ser.close()
        return True

    def protocol_server_to_dfu(self, serial_port, protocol_id):
        ser = self._try_open_port(serial_port)
        ser.write(bytes([0xCC, 0x02, 0x00, protocol_id, 0xFF, 0xFF, 0xCD]))
        ser.close()
        return True

    @abc.abstractmethod
    def device_enter_dfu(self, device_pid, device_port):
        """Implemented in <device>_dfu_config.py"""

    def enter_dfu(self, device_port):
        """Function to make the devices enter Bootloader/DFU mode"""
        if isinstance(device_port, SerialDevice):
            device_pid = self._get_pid_from_port(device_port.port)
            serial_port_name = device_port.port
        else:
            device_pid = device_port.pid
            serial_port_name = None

        if not self.device_enter_dfu(device_pid, serial_port_name):
            msg = "Device not found"
            raise FlashException(msg)

        # Wait for DFU device to appear
        retries = 5
        device_found = False
        cdc_port = None
        while retries > 0:
            log.debug("Wait for DFU device...")
            time.sleep(1)
            cdc_port = BootloaderTool._find_port(self.device_vid, self.bootloader_pid, device_port)
            if cdc_port is not None:
                device_found = True
                break
            retries -= 1

        if not device_found:
            msg = "DFU device not found"
            raise FlashException(msg)

    @staticmethod
    def _try_open_port(serial_port):
        retries = 5
        ser = None
        while retries > 0 and ser is None:
            try:
                ser = serial.Serial(serial_port, exclusive=True)
            except SerialException:
                ser = None
                time.sleep(0.2)
                retries -= 1

        if ser is None:
            msg = f"Flash failed, {serial_port} cannot be opened"
            raise FlashException(msg)

        return ser

    def flash(self, device_port, device_name, image_path, progress_callback=None):
        """Flash an firmware image to the device"""
        self.enter_dfu(device_port)
        dfu_port = self._find_port(self.device_vid, self.bootloader_pid, device_port)
        serial_dev = None
        if dfu_port is not None:
            serial_dev = self._try_open_port(dfu_port)
        else:
            msg = "Flash failed, device not found"
            raise FlashException(msg)

        try:
            with ImageFlasher(serial_dev) as image_flasher:
                image_flasher.set_progress_callback(progress_callback)
                image_flasher.flash_image_file(image_path)
        except CommandFailed as exc:
            if "Data does not contain a valid image" in str(exc):
                error_string = (
                    f"ERROR: The selected file does not contain a valid image for {device_name}"
                )
            else:
                error_string = f"ERROR: {str(exc)}"
            raise FlashException(error_string)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from acconeer.exptool import a121

from ._processor import MeasurementType, ProcessorConfig


def get_close_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        sweeps_per_frame=16,
        sweep_rate=320,
        continuous_sweep_mode=True,
        double_buffering=True,
        inter_sweep_idle_state=a121.IdleState.READY,
        inter_frame_idle_state=a121.IdleState.READY,
        num_points=3,
        profile=a121.Profile.PROFILE_1,
        receiver_gain=0,
        hwaas=40,
        start_point=0,
        step_length=6,
    )


def get_close_processor_config() -> ProcessorConfig:
    return ProcessorConfig(measurement_type=MeasurementType.CLOSE_RANGE)


def get_far_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        sweeps_per_frame=16,
        sweep_rate=320,
        continuous_sweep_mode=True,
        double_buffering=True,
        inter_sweep_idle_state=a121.IdleState.READY,
        inter_frame_idle_state=a121.IdleState.READY,
        start_point=0,
        num_points=3,
        step_length=24,
        profile=a121.Profile.PROFILE_3,
        hwaas=60,
        receiver_gain=5,
    )


def get_far_processor_config() -> ProcessorConfig:
    return ProcessorConfig(measurement_type=MeasurementType.FAR_RANGE)


def get_close_and_far_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        sweeps_per_frame=16,
        sweep_rate=320,
        continuous_sweep_mode=True,
        double_buffering=True,
        inter_sweep_idle_state=a121.IdleState.READY,
        inter_frame_idle_state=a121.IdleState.READY,
        subsweeps=[
            a121.SubsweepConfig(
                start_point=0,
                num_points=3,
                step_length=6,
                profile=a121.Profile.PROFILE_1,
                hwaas=40,
                receiver_gain=0,
            ),
            a121.SubsweepConfig(
                start_point=0,
                num_points=3,
                step_length=24,
                profile=a121.Profile.PROFILE_3,
                hwaas=60,
                receiver_gain=5,
            ),
        ],
    )


def get_close_and_far_processor_config() -> ProcessorConfig:
    return ProcessorConfig(measurement_type=MeasurementType.CLOSE_AND_FAR_RANGE)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import copy
import functools
import itertools
import typing as t

import typing_extensions as te

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QTabWidget, QVBoxLayout, QWidget

import pyqtgraph as pg

from acconeer.exptool import a121
from acconeer.exptool.a121._core import utils as core_utils
from acconeer.exptool.a121.model import power
from acconeer.exptool.app.new.ui.icons import WARNING_YELLOW
from acconeer.exptool.app.new.ui.resource_tab.event_system import (
    ChangeIdEvent,
    EventBroker,
    IdentifiedServiceUninstalledEvent,
)
from acconeer.exptool.utils import pg_pen_cycler

from .distance_config_input import DistanceConfigEvent
from .presence_config_input import PresenceConfigEvent
from .session_config_input import SessionConfigEvent


_T = t.TypeVar("_T")


def incremental_plot(
    xs: list[float],
    f: t.Callable[[float], float],
    ordering_strategy: t.Callable[[int], t.Iterable[int]],
) -> t.Iterator[tuple[list[float], list[float]]]:
    """Generator that adds one (x, y)-point to the plot each time it's iterated"""
    done: list[tuple[float, float]] = []

    for i in ordering_strategy(len(xs)):
        x = xs[i]

        done += [(x, f(x))]
        done.sort(key=lambda t: t[0])

        all_xs = [done_x for done_x, _ in done]
        all_ys = [done_y for _, done_y in done]
        yield (all_xs, all_ys)


class _PowerConsumptionVsRatePlot(pg.PlotWidget):
    def __init__(self, algorithm: power.algo.Algorithm) -> None:
        super().__init__()

        self._algorithm: te.Final[power.algo.Algorithm] = algorithm

        self.getPlotItem().setLabel("bottom", "Update rate", units="Hz")
        self.getPlotItem().setLabel("left", "Sensor + XM125", units="A")
        self.getPlotItem().setContentsMargins(0, 0, 0, 10)
        self.getPlotItem().addLegend()
        self.getViewBox().setMouseMode(pg.ViewBox.PanMode)

        self._ready_plotter: t.Iterator[tuple[list[float], list[float]]] = iter([])
        self._sleep_plotter: t.Iterator[tuple[list[float], list[float]]] = iter([])
        self._deep_sleep_plotter: t.Iterator[tuple[list[float], list[float]]] = iter([])
        self._hibernate_plotter: t.Iterator[tuple[list[float], list[float]]] = iter([])
        self._off_plotter: t.Iterator[tuple[list[float], list[float]]] = iter([])

        self._ready_curve: t.Optional[pg.PlotDataItem] = None
        self._sleep_curve: t.Optional[pg.PlotDataItem] = None
        self._deep_sleep_curve: t.Optional[pg.PlotDataItem] = None
        self._hibernate_curve: t.Optional[pg.PlotDataItem] = None
        self._off_curve: t.Optional[pg.PlotDataItem] = None

        self._plot_increment_timer = QTimer()
        self._plot_increment_timer.timeout.connect(self._increment_plots)

    @staticmethod
    def _get_update_rates(update_rate: float) -> list[float]:
        return list(
            itertools.takewhile(
                lambda e: e < update_rate * 1.5,
                itertools.chain(
                    [i / 100 for i in range(1, 100)],  # [0.01, 0.99] in 0.01 steps
                    [1 + i / 10 for i in range(90)],  # [1, 9.9] in 0.1 steps
                    itertools.count(start=10),  # [10, inf) in 1.0 steps
                ),
            )
        )

    @staticmethod
    def _evolve_config(
        config: a121.SessionConfig,
        *,
        update_rate: t.Optional[float] = None,
        inter_frame_idle_state: t.Optional[a121.IdleState] = None,
    ) -> a121.SessionConfig:
        config_copy = copy.deepcopy(config)

        if update_rate is not None:
            config_copy.update_rate = update_rate

        if inter_frame_idle_state is not None:
            last_sensor_config = list(config_copy.groups[-1].values())[-1]
            last_sensor_config.inter_frame_idle_state = inter_frame_idle_state

        return config_copy

    def _increment_plots(self) -> None:
        (self._sleep_curve, sleep_done) = self._increment_plot(
            "Sleep", pg_pen_cycler(0), self._sleep_plotter, self._sleep_curve
        )
        (self._deep_sleep_curve, deep_sleep_done) = self._increment_plot(
            "Deep sleep", pg_pen_cycler(1), self._deep_sleep_plotter, self._deep_sleep_curve
        )
        (self._hibernate_curve, hibernate_done) = self._increment_plot(
            "Hibernate", pg_pen_cycler(2), self._hibernate_plotter, self._hibernate_curve
        )
        (self._off_curve, off_done) = self._increment_plot(
            "Off", pg_pen_cycler(3), self._off_plotter, self._off_curve
        )
        (self._ready_curve, _) = self._increment_plot(
            "Ready", pg_pen_cycler(4), self._ready_plotter, self._ready_curve
        )

        if all([sleep_done, deep_sleep_done, hibernate_done, off_done]):
            self._plot_increment_timer.stop()

    def _increment_plot(
        self,
        name: str,
        pen: t.Any,
        incremental_plotter: t.Iterator[tuple[list[float], list[float]]],
        reused_curve: t.Optional[pg.PlotDataItem],
    ) -> tuple[pg.PlotDataItem, bool]:
        """
        Plots the next increment of incremental_plotter.
        If reused_curve is None, a new curve will be created with name and pen.

        Returns the updated curve (migth be the same as reused_curve) and whether
        plotting is completed
        """
        try:
            (xs, ys) = next(incremental_plotter)
        except StopIteration:
            return (reused_curve, True)
        else:
            if reused_curve is None:
                return (
                    self.plot(xs, ys, name=name, pen=pen),
                    False,
                )
            else:
                reused_curve.setData(xs, ys)
                return (reused_curve, False)

    @staticmethod
    def _will_keep_rate(
        config: a121.SessionConfig,
        lower_idle_state: t.Optional[power.Sensor.LowerIdleState],
        algorithm: power.algo.Algorithm,
    ) -> bool:
        rate = power.configured_rate(config)

        if rate is None:
            return True

        active = power.group_active(config, lower_idle_state, algorithm)
        return active.duration < 1 / rate

    def update_power_curves(
        self,
        config: a121.SessionConfig,
        lower_idle_state: t.Optional[power.Sensor.LowerIdleState],
    ) -> None:
        self._plot_increment_timer.stop()
        self.clear()

        configured_rate = power.configured_rate(config)

        if configured_rate is None:
            self.disableAutoRange()
            ((_, x_max), (_, _)) = self.viewRange()

            text_item = pg.InfiniteLine(
                pos=0.75 * x_max,
                angle=0,
                label="Set update rate to compare\npower states over rates",
                labelOpts={"color": "#111", "fill": "#ddd"},
                pen=pg.mkPen(None),
            )
            self.addItem(text_item)
            return

        update_rates = self._get_update_rates(configured_rate)

        self.enableAutoRange()
        self.setXRange(min(update_rates), max(update_rates))

        self._ready_curve = None
        self._sleep_curve = None
        self._deep_sleep_curve = None
        self._hibernate_curve = None
        self._off_curve = None
        self._increment_plots()
        self.addItem(
            pg.ScatterPlotItem(
                [configured_rate],
                [
                    power.converged_average_current(
                        config,
                        lower_idle_state=lower_idle_state,
                        absolute_tolerance=1e-3,
                        algorithm=self._algorithm,
                    )
                ],
                name="Current config",
            )
        )

        curves_that_wont_keep_rate = []
        if any(
            sensor_config.inter_frame_idle_state == a121.IdleState.READY
            for sensor_config in core_utils.iterate_extended_structure_values(config.groups)
        ):
            ready_config_evolver = functools.partial(
                self._evolve_config,
                inter_frame_idle_state=a121.IdleState.READY,
            )
            if not self._will_keep_rate(
                ready_config_evolver(config, update_rate=max(update_rates)),
                lower_idle_state=None,
                algorithm=self._algorithm,
            ):
                curves_that_wont_keep_rate += ["Ready"]

            def ready_f(update_rate: float) -> float:
                return power.converged_average_current(
                    ready_config_evolver(config, update_rate=update_rate),
                    lower_idle_state=None,
                    absolute_tolerance=1e-3,
                    algorithm=self._algorithm,
                )

            self._ready_plotter = incremental_plot(update_rates, ready_f, ordering_strategy=range)

        sleep_config_evolver = functools.partial(
            self._evolve_config,
            inter_frame_idle_state=a121.IdleState.SLEEP,
        )
        if not self._will_keep_rate(
            sleep_config_evolver(config, update_rate=max(update_rates)),
            lower_idle_state=None,
            algorithm=self._algorithm,
        ):
            curves_that_wont_keep_rate += ["Sleep"]

        def sleep_f(update_rate: float) -> float:
            return power.converged_average_current(
                sleep_config_evolver(config, update_rate=update_rate),
                lower_idle_state=None,
                absolute_tolerance=1e-3,
                algorithm=self._algorithm,
            )

        self._sleep_plotter = incremental_plot(update_rates, sleep_f, ordering_strategy=range)

        deep_sleep_config_evolver = functools.partial(
            self._evolve_config,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        )
        if not self._will_keep_rate(
            deep_sleep_config_evolver(config, update_rate=max(update_rates)),
            lower_idle_state=None,
            algorithm=self._algorithm,
        ):
            curves_that_wont_keep_rate += ["Deep Sleep"]

        def deep_sleep_f(update_rate: float) -> float:
            return power.converged_average_current(
                deep_sleep_config_evolver(config, update_rate=update_rate),
                lower_idle_state=None,
                absolute_tolerance=1e-3,
                algorithm=self._algorithm,
            )

        self._deep_sleep_plotter = incremental_plot(
            update_rates, deep_sleep_f, ordering_strategy=range
        )

        if not self._will_keep_rate(
            self._evolve_config(config, update_rate=max(update_rates)),
            lower_idle_state=power.Sensor.IdleState.HIBERNATE,
            algorithm=self._algorithm,
        ):
            curves_that_wont_keep_rate += ["Hibernate"]

        def hibernate_f(update_rate: float) -> float:
            return power.converged_average_current(
                self._evolve_config(config, update_rate=update_rate),
                lower_idle_state=power.Sensor.IdleState.HIBERNATE,
                absolute_tolerance=1e-3,
                algorithm=self._algorithm,
            )

        self._hibernate_plotter = incremental_plot(
            update_rates, hibernate_f, ordering_strategy=range
        )

        if not self._will_keep_rate(
            self._evolve_config(config, update_rate=max(update_rates)),
            lower_idle_state=power.Sensor.IdleState.OFF,
            algorithm=self._algorithm,
        ):
            curves_that_wont_keep_rate += ["Off"]

        def off_f(update_rate: float) -> float:
            return power.converged_average_current(
                self._evolve_config(config, update_rate=update_rate),
                lower_idle_state=power.Sensor.IdleState.OFF,
                absolute_tolerance=1e-3,
                algorithm=self._algorithm,
            )

        self._off_plotter = incremental_plot(update_rates, off_f, ordering_strategy=range)

        if curves_that_wont_keep_rate:
            rate_warning_text = pg.InfiniteLine(
                pos=0.003,
                angle=0,
                label=(
                    "The following curves cannot keep\n"
                    + "rate at some point in the plot:\n\n"
                    + "\n".join(f" - {name}" for name in curves_that_wont_keep_rate)
                ),
                labelOpts={
                    "color": "#0008",
                    "fill": pg.mkBrush(WARNING_YELLOW + "88"),
                    "border": "#0008",
                },
                pen=pg.mkPen(None),
            )
            self.addItem(rate_warning_text)

        self._plot_increment_timer.start(10)


class PowerConsumptionVsRateOutput(QWidget):
    INTERESTS: t.ClassVar[set[type]] = {
        SessionConfigEvent,
        IdentifiedServiceUninstalledEvent,
        DistanceConfigEvent,
        PresenceConfigEvent,
        ChangeIdEvent,
    }
    description: t.ClassVar[str] = "\n\n".join(
        [
            "Compares power consumption and rate for different power- & inter frame idle states.",
            "NOTE: A curve flatlining means that the rate is the maximum for that power state.",
        ]
    )
    window_title = "Power consumption vs rate"

    def __init__(
        self,
        broker: EventBroker,
    ) -> None:
        super().__init__()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        self.setStyleSheet("QTabBar { font: bold 14px; font-family: monospace; }")
        self._tab_widget = QTabWidget()
        self._tabs: dict[str, _PowerConsumptionVsRatePlot] = {}

        layout.addWidget(self._tab_widget)
        self.setLayout(layout)

        broker.install_service(self)
        self.uninstall_function = lambda: broker.uninstall_service(self)
        broker.brief_service(self)

    def handle_event(self, event: t.Any) -> None:
        if isinstance(event, SessionConfigEvent):
            self._handle_session_config_event(event)
        elif isinstance(event, DistanceConfigEvent):
            self._handle_distance_config_event(event)
        elif isinstance(event, PresenceConfigEvent):
            self._handle_presence_config_event(event)
        elif isinstance(event, IdentifiedServiceUninstalledEvent):
            self._handle_identified_service_uninstalled_event(event)
        elif isinstance(event, ChangeIdEvent):
            self._handle_change_id_event(event)
        else:
            raise NotImplementedError

    def _handle_session_config_event(self, event: SessionConfigEvent) -> None:
        if event.service_id not in self._tabs:
            plot_widget = _PowerConsumptionVsRatePlot(algorithm=power.algo.SparseIq())
            self._tabs[event.service_id] = plot_widget

            self._tab_widget.addTab(plot_widget, event.service_id)
            self._tab_widget.setCurrentWidget(plot_widget)

        self._tabs[event.service_id].update_power_curves(
            event.session_config, event.lower_idle_state
        )

    def _handle_distance_config_event(self, event: DistanceConfigEvent) -> None:
        if event.service_id not in self._tabs:
            plot_widget = _PowerConsumptionVsRatePlot(algorithm=power.algo.Distance())
            self._tabs[event.service_id] = plot_widget

            self._tab_widget.addTab(plot_widget, event.service_id)
            self._tab_widget.setCurrentWidget(plot_widget)

        self._tabs[event.service_id].update_power_curves(
            event.translated_session_config, event.lower_idle_state
        )

    def _handle_presence_config_event(self, event: PresenceConfigEvent) -> None:
        if event.service_id not in self._tabs:
            plot_widget = _PowerConsumptionVsRatePlot(algorithm=power.algo.Presence())
            self._tabs[event.service_id] = plot_widget

            self._tab_widget.addTab(plot_widget, event.service_id)
            self._tab_widget.setCurrentWidget(plot_widget)

        self._tabs[event.service_id].update_power_curves(
            event.translated_session_config, event.lower_idle_state
        )

    def _handle_identified_service_uninstalled_event(
        self, event: IdentifiedServiceUninstalledEvent
    ) -> None:
        tab_widget = self._tabs.pop(event.id_)
        tab_index = self._tab_widget.indexOf(tab_widget)
        if tab_index != -1:
            self._tab_widget.removeTab(tab_index)

    def _handle_change_id_event(self, event: ChangeIdEvent) -> None:
        tab_widget = self._tabs.pop(event.old_id, None)
        if tab_widget is not None:
            tab_index = self._tab_widget.indexOf(tab_widget)
            self._tab_widget.setTabText(tab_index, event.new_id)
            self._tabs[event.new_id] = tab_widget



# Copyright (c) Acconeer AB, 2024
# All rights reserved

from acconeer.exptool import a121

from ._processor import ProcessorConfig


def get_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        subsweeps=[
            a121.SubsweepConfig(
                start_point=40,
                step_length=8,
                num_points=14,
                profile=a121.Profile.PROFILE_1,
                hwaas=4,
            ),
            a121.SubsweepConfig(
                start_point=150,
                step_length=12,
                num_points=23,
                profile=a121.Profile.PROFILE_3,
                hwaas=8,
            ),
        ],
        sweeps_per_frame=32,
        frame_rate=5,
    )


def get_processor_config() -> ProcessorConfig:
    return ProcessorConfig()



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoProcessorConfigBase
from acconeer.exptool.a121.algo.presence import DetectorConfig, DetectorMetadata, DetectorResult


@attrs.frozen(kw_only=True)
class ProcessorResult:
    """Processor result"""

    zone_limits: npt.NDArray[np.float64] = attrs.field()
    max_presence_zone: Optional[int] = attrs.field()
    total_zone_detections: npt.NDArray[np.int_] = attrs.field()
    inter_zone_detections: npt.NDArray[np.int_] = attrs.field()
    max_inter_zone: Optional[int] = attrs.field()
    intra_zone_detections: npt.NDArray[np.int_] = attrs.field()
    max_intra_zone: Optional[int] = attrs.field()


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    num_zones: int = attrs.field(default=3)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        return []


class Processor:
    """Smart presence processor
    :param processor_config: Processor configuration.
    """

    max_inter_zone: Optional[int]
    max_intra_zone: Optional[int]

    def __init__(
        self,
        processor_config: ProcessorConfig,
        detector_config: DetectorConfig,
        session_config: a121.SessionConfig,
        detector_metadata: DetectorMetadata,
    ):
        num_points = sum(
            [subsweep.num_points for subsweep in session_config.sensor_config.subsweeps]
        )
        self.num_zones = np.minimum(processor_config.num_zones, num_points)
        self.distances = np.linspace(
            detector_metadata.start_m,
            detector_metadata.end_m,
            num_points,
        )
        self.zone_limits = self.create_zones(self.distances, self.num_zones)

        self.inter_enable = detector_config.inter_enable
        self.inter_threshold = detector_config.inter_detection_threshold
        self.inter_zones = np.zeros(self.num_zones, dtype=int)
        self.max_inter_zone = None

        self.intra_enable = detector_config.intra_enable
        self.intra_threshold = detector_config.intra_detection_threshold
        self.intra_zones = np.zeros(self.num_zones, dtype=int)
        self.max_intra_zone = None

    @staticmethod
    def create_zones(
        distances: npt.NDArray[np.float64], num_zones: int
    ) -> npt.NDArray[np.float64]:
        """Create zones limits based on chosen range and number of zones."""

        if num_zones < distances.size:
            zone_limits = np.linspace(distances[0], distances[-1], num_zones + 1)
            # Only return upper limits
            return zone_limits[1:]  # type: ignore[no-any-return]

        else:
            return distances

    def get_zone_detections(
        self, depthwise_scores: npt.NDArray[np.float64], threshold: float
    ) -> npt.NDArray[np.int_]:
        """Get presence detection result for all zones."""

        zones = np.zeros(self.num_zones, dtype=int)
        zone_detected = False
        limit_idx = 0
        for i, score in enumerate(depthwise_scores):
            if self.distances[i] > self.zone_limits[limit_idx]:
                limit_idx += 1
                zone_detected = False
            if not zone_detected:
                zone_detected = score > threshold
                zones[limit_idx] = 1 if zone_detected else 0

        return zones

    def get_max_presence_zone(self, depthwise_scores: npt.NDArray[np.float64]) -> int:
        """Get the zone with maximum presence score."""

        max_idx = np.argmax(depthwise_scores)
        max_distance = self.distances[max_idx]
        max_zone = 0
        while max_zone < self.num_zones - 1 and max_distance > self.zone_limits[max_zone]:
            max_zone += 1

        return max_zone

    def process(self, result: DetectorResult) -> ProcessorResult:
        max_presence_zone: Optional[int]

        if result.presence_detected:
            # Update zone detections for each detection type.
            if self.inter_enable:
                if result.inter_presence_score > self.inter_threshold:
                    self.inter_zones = self.get_zone_detections(
                        result.inter_depthwise_scores, self.inter_threshold
                    )
                    if not np.any(self.inter_zones):
                        self.inter_zones[self.max_inter_zone] = 1
                    else:
                        self.max_inter_zone = self.get_max_presence_zone(
                            result.inter_depthwise_scores
                        )

                else:
                    self.inter_zones = np.zeros(self.num_zones, dtype=int)
                    self.max_inter_zone = None

            if self.intra_enable:
                if result.intra_presence_score > self.intra_threshold:
                    self.intra_zones = self.get_zone_detections(
                        result.intra_depthwise_scores, self.intra_threshold
                    )
                    if not np.any(self.intra_zones):
                        self.intra_zones[self.max_intra_zone] = 1
                    else:
                        self.max_intra_zone = self.get_max_presence_zone(
                            result.intra_depthwise_scores
                        )

                else:
                    self.intra_zones = np.zeros(self.num_zones, dtype=int)
                    self.max_intra_zone = None

            # max intra zone is prioritized due to faster reaction time.
            if self.max_intra_zone is not None:
                max_presence_zone = self.max_intra_zone
            else:
                max_presence_zone = self.max_inter_zone

            total_zone_detections = np.maximum(self.inter_zones, self.intra_zones)

        else:
            max_presence_zone = None
            self.inter_zones = np.zeros(self.num_zones, dtype=int)
            self.intra_zones = np.zeros(self.num_zones, dtype=int)
            total_zone_detections = np.zeros(self.num_zones, dtype=int)
            self.max_inter_zone = None
            self.max_intra_zone = None

        return ProcessorResult(
            zone_limits=self.zone_limits,
            max_presence_zone=max_presence_zone,
            total_zone_detections=total_zone_detections,
            inter_zone_detections=self.inter_zones,
            max_inter_zone=self.max_inter_zone,
            intra_zone_detections=self.intra_zones,
            max_intra_zone=self.max_intra_zone,
        )



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

"""
Pidget- & PidgetGroup hooks

This module defines closures and higher order functions that
allows declarative definition of Pidget- & PidgetGroup hooks.

It's still possible to use any function that adheres to respective type.

The definitions of these hooks are placed in ``pidget_groups.py`` and
``pidgets.py``, respectively.

.. code-block:: python
    WidgetHook = t.Callable[
        [QWidget, t.Mapping[str, Pidget]], None
    ]
    PidgetHook = Callable[
        ["Pidget", Mapping[str, "Pidget"]], None
    ]

A ``WidgetHook`` is a function that accepts a ``QWidget`` instance
and a ``Mapping[str, Pidget]`` and returns nothing.
The ``QWidget`` instance is the container widget returned by the
``get_container`` function of the ``PidgetGroup``.

.. code-block:: python
    def my_hook(pg: QWidget, mapping: Mapping[str, Pidget]) -> None:
        print(pg)  # pg will be the "a" variable below

    a = PidgetGroup(
        hooks=my_hook,
        # or hooks=[my_hook]
        # or hooks=(my_hook,)
    )

The second argument of a ``WidgetHook`` is the ``PidgetMapping`` kept
internally in ``AttrsConfigEditor``;

.. code-block:: python
    def my_hook(pg: QWidget, mapping: Mapping[str, Pidget]) -> None:
        print(mapping)
        # mapping will be:
        # {
        #     # Note: these are the actual pidgets!
        #     "int_parameter": IntPidget()
        #     "float_parameter": FloatPidget()
        # }
        #

    mapping: PidgetGroupFactoryMapping = {
        PidgetGroup(hooks=my_hook): {
            "int_parameter": IntPidgetFactory(...),
            "float_parameter": FloatPidgetFactory(...),
        }
    }

P.S. The functionality is analogous in ``PidgetHook``, but the
actual ``Pidget`` the hook is attached to is the first argument of the hook.


This means that the hook has access to any pidget in a ``AttrsConfigEditor``,
allowing for interesting combinations!
"""

from __future__ import annotations

import typing as t

from PySide6.QtWidgets import QWidget

from .pidget_groups import WidgetHook
from .pidgets import Pidget


PidgetMapping = t.Mapping[str, Pidget]

PidgetMappingPredicate = t.Callable[[PidgetMapping], bool]


def parameter_equals(aspect: str, value: t.Any) -> PidgetMappingPredicate:
    """Checks whether the parameter of the pidget assigned to aspect is equal to value"""

    def inner(mapping: PidgetMapping) -> bool:
        return bool(mapping[aspect].get_data() == value)

    return inner


def parameter_is(aspect: str, value: t.Any) -> PidgetMappingPredicate:
    """Checks whether the parameter of the pidget assigned to aspect "is" passed value
    Should be used with singletons like "True", "False", "None", etc.
    """

    def inner(mapping: PidgetMapping) -> bool:
        return mapping[aspect].get_data() is value

    return inner


def parameter_within(aspect: str, range: t.Tuple[float, float]) -> PidgetMappingPredicate:
    """Checks whether the parameter of the pidget assigned to aspect is in range"""

    def inner(mapping: PidgetMapping) -> bool:
        lower, higher = range
        return bool(lower < mapping[aspect].get_data() < higher)

    return inner


def enable_if(*predicates: PidgetMappingPredicate) -> WidgetHook:
    """Enables the instance this hook is assigned to if all the predicates evaluates to "True" """

    def inner(inst: QWidget, mapping: PidgetMapping) -> None:
        inst.setEnabled(all(predicate(mapping) for predicate in predicates))

    return inner


def disable_if(*predicates: PidgetMappingPredicate) -> WidgetHook:
    """Negation of "enable_if" (all predicates should evaluate to False)"""

    def inner(inst: QWidget, mapping: PidgetMapping) -> None:
        inst.setEnabled(not any(predicate(mapping) for predicate in predicates))

    return inner



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import typing as t

from .pidgets import PidgetFactory, PidgetGroup


PidgetFactoryMapping = t.Mapping[str, PidgetFactory]
PidgetGroupFactoryMapping = t.Mapping[PidgetGroup, PidgetFactoryMapping]



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import json
import warnings
from typing import Any, List, Optional, TypeVar

import numpy as np

from acconeer.exptool import opser
from acconeer.exptool._core.class_creation.descriptors import Descriptor, delegate_field
from acconeer.exptool._core.class_creation.formatting import pretty_dict_line_strs
from acconeer.exptool._core.entities.validation_result import (
    ValidationError,
    ValidationResult,
    ValidationWarning,
)
from acconeer.exptool.a121._core import utils

from .config_enums import PRF, IdleState, Profile
from .subsweep_config import SubsweepConfig


T = TypeVar("T")

_TOO_MANY_SUBSWEEPS_ERROR_FORMAT = (
    "SensorConfig has too many subsweeps "
    + 'to use accessor "{}". Number of subsweeps needs to be 1.'
)
VALIDATION_TAG_BUFFER_SIZE_TOO_LARGE = object()


def subsweep_delegate_field(descriptor: Any, type_: type[T]) -> Descriptor[T]:
    return delegate_field(
        "subsweep",
        descriptor,
        type_=type_,
        doc=SubsweepConfig,
        error_format=_TOO_MANY_SUBSWEEPS_ERROR_FORMAT.format("{descriptor_name}"),
    )


@utils.no_dynamic_member_creation
class SensorConfig:
    """Sensor configuration

    The sensor config represents a 1-1 mapping to the RSS service config.

    By default, the sensor config holds a single :class:`SubsweepConfig`. The parameters defined by
    the subsweep config, like :attr:`start_point`, can be accessed via the sensor config. If
    multiple subsweeps are used, those parameters must be accessed via their respective subsweep
    configs.

    For example, a sensor config can be created like this:

    .. code-block:: python

        SensorConfig(sweeps_per_frame=16, start_point=123)

    Note that the :attr:`start_point` is implicitly set in the underlying subsweep config. If you
    want to explicitly set the subsweep config(s), you can do

    .. code-block:: python

        SensorConfig(
            sweeps_per_frame=16,
            subsweeps=[
                SubsweepConfig(start_point=123),
            ],
        )

    Parameters can also be accessed via the class attributes:

    .. code-block:: python

        sensor_config = SensorConfig()
        sensor_config.sweeps_per_frame = 16
        sensor_config.start_point = 123

    If you want to use multiple subsweeps with this style of setting/getting the attributes, you
    can do like this:

    .. code-block:: python

        sensor_config = SensorConfig(num_subsweeps=3)
        sensor_config.sweeps_per_frame = 16
        sensor_config.subsweeps[0].start_point = 123

    .. note::

        The sensor config does not control on which sensor it should be run. That is handled by
        the :class:`SessionConfig`.

    :param subsweeps:
        The list of subsweeps to initialize with. May not be combined with ``num_subsweeps``.
    :param num_subsweeps:
        Initialize with a given number of subsweeps. May not be combined with ``subsweeps``.
    :raises ValueError: If ``subsweeps`` and ``num_subsweeps`` are both given.
    :raises ValueError: If the given list of ``subsweeps`` is empty.
    :raises ValueError: If subsweeps parameters are both given implicitly and via ``subsweeps``.
    """

    MAX_HWAAS = SubsweepConfig.MAX_HWAAS

    _subsweeps: List[SubsweepConfig]

    _sweeps_per_frame: int
    _sweep_rate: Optional[float]
    _frame_rate: Optional[float]
    _continuous_sweep_mode: bool
    _double_buffering: bool
    _inter_frame_idle_state: IdleState
    _inter_sweep_idle_state: IdleState

    # Seems like there is a false positive where mypy confuses
    # the below descriptors (class members) of attrs.fields with __slots__ variables.
    start_point = subsweep_delegate_field(SubsweepConfig.start_point, type_=int)  # type: ignore[misc]
    num_points = subsweep_delegate_field(SubsweepConfig.num_points, type_=int)  # type: ignore[misc]
    step_length = subsweep_delegate_field(SubsweepConfig.step_length, type_=int)  # type: ignore[misc]
    profile = subsweep_delegate_field(SubsweepConfig.profile, type_=Profile)  # type: ignore[misc]
    hwaas = subsweep_delegate_field(SubsweepConfig.hwaas, type_=int)  # type: ignore[misc]
    receiver_gain = subsweep_delegate_field(SubsweepConfig.receiver_gain, type_=int)  # type: ignore[misc]
    enable_tx = subsweep_delegate_field(SubsweepConfig.enable_tx, type_=bool)  # type: ignore[misc]
    enable_loopback = subsweep_delegate_field(SubsweepConfig.enable_loopback, type_=bool)  # type: ignore[misc]
    phase_enhancement = subsweep_delegate_field(SubsweepConfig.phase_enhancement, type_=bool)  # type: ignore[misc]
    iq_imbalance_compensation = subsweep_delegate_field(
        SubsweepConfig.iq_imbalance_compensation,  # type: ignore[misc]
        type_=bool,
    )
    prf = subsweep_delegate_field(SubsweepConfig.prf, type_=PRF)

    def __init__(
        self,
        *,
        subsweeps: Optional[list[SubsweepConfig]] = None,
        num_subsweeps: Optional[int] = None,
        sweeps_per_frame: int = 1,
        sweep_rate: Optional[float] = None,
        frame_rate: Optional[float] = None,
        continuous_sweep_mode: bool = False,
        double_buffering: bool = False,
        inter_frame_idle_state: IdleState = IdleState.DEEP_SLEEP,
        inter_sweep_idle_state: IdleState = IdleState.READY,
        start_point: Optional[int] = None,
        num_points: Optional[int] = None,
        step_length: Optional[int] = None,
        profile: Optional[Profile] = None,
        hwaas: Optional[int] = None,
        receiver_gain: Optional[int] = None,
        enable_tx: Optional[bool] = None,
        enable_loopback: Optional[bool] = None,
        phase_enhancement: Optional[bool] = None,
        iq_imbalance_compensation: Optional[bool] = None,
        prf: Optional[PRF] = None,
    ) -> None:
        if subsweeps is not None and num_subsweeps is not None:
            msg = "It is not allowed to pass both `subsweeps` and `num_subsweeps`. Choose one."
            raise ValueError(msg)
        if subsweeps == []:
            msg = "Cannot pass an empty `subsweeps` list."
            raise ValueError(msg)

        no_subsweep_param_passed = (
            start_point is None
            and num_points is None
            and step_length is None
            and profile is None
            and hwaas is None
            and receiver_gain is None
            and enable_tx is None
            and enable_loopback is None
            and phase_enhancement is None
            and iq_imbalance_compensation is None
            and prf is None
        )
        if subsweeps is not None and not no_subsweep_param_passed:
            raise ValueError(
                "Combining 'subsweeps' and subsweep parameters is not allowed."
                + "Specify subsweep params in each 'SubsweepConfig' instead."
            )

        if subsweeps is None and num_subsweeps is None:
            num_subsweeps = 1

        if subsweeps is not None:
            self._subsweeps = subsweeps
        elif num_subsweeps is not None:
            self._subsweeps = [SubsweepConfig() for _ in range(num_subsweeps)]
        else:
            raise RuntimeError

        self.sweeps_per_frame = sweeps_per_frame
        self.sweep_rate = sweep_rate
        self.frame_rate = frame_rate
        self.continuous_sweep_mode = continuous_sweep_mode
        self.double_buffering = double_buffering
        self.inter_frame_idle_state = inter_frame_idle_state
        self.inter_sweep_idle_state = inter_sweep_idle_state

        # Init proxy attributes

        if hwaas is not None:
            self.hwaas = hwaas
        if start_point is not None:
            self.start_point = start_point
        if num_points is not None:
            self.num_points = num_points
        if step_length is not None:
            self.step_length = step_length
        if profile is not None:
            self.profile = profile
        if receiver_gain is not None:
            self.receiver_gain = receiver_gain
        if enable_tx is not None:
            self.enable_tx = enable_tx
        if enable_loopback is not None:
            self.enable_loopback = enable_loopback
        if phase_enhancement is not None:
            self.phase_enhancement = phase_enhancement
        if iq_imbalance_compensation is not None:
            self.iq_imbalance_compensation = iq_imbalance_compensation
        if prf is not None:
            self.prf = prf

    @property
    def subsweep(self) -> SubsweepConfig:
        """Retrieves the sole ``SubsweepConfig``

        :raises AttributeError: If ``num_subsweeps`` > 1
        """
        if self.num_subsweeps > 1:
            raise AttributeError(_TOO_MANY_SUBSWEEPS_ERROR_FORMAT.format("subsweep"))

        return self.subsweeps[0]

    @property
    def subsweeps(self) -> list[SubsweepConfig]:
        """The list of subsweep configs"""

        return self._subsweeps

    @property
    def num_subsweeps(self) -> int:
        """The number of subsweep configs"""

        return len(self.subsweeps)

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.to_dict() == other.to_dict()

    def to_dict(self) -> dict[str, Any]:
        return {
            "sweep_rate": self.sweep_rate,
            "frame_rate": self.frame_rate,
            "continuous_sweep_mode": self.continuous_sweep_mode,
            "double_buffering": self.double_buffering,
            "inter_frame_idle_state": self.inter_frame_idle_state,
            "inter_sweep_idle_state": self.inter_sweep_idle_state,
            "sweeps_per_frame": self.sweeps_per_frame,
            "subsweeps": [subsweep.to_dict() for subsweep in self.subsweeps],
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> SensorConfig:
        d = d.copy()
        d["subsweeps"] = [SubsweepConfig.from_dict(subsweep_d) for subsweep_d in d["subsweeps"]]
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), cls=utils.EntityJSONEncoder)

    @classmethod
    def from_json(cls, json_str: str) -> SensorConfig:
        return cls.from_dict(json.loads(json_str))

    def validate(self) -> None:
        """Performs self-validation and validation of its subsweep configs

        :raises ValidationError: If anything is invalid.
        """
        for validation_result in self._collect_validation_results():
            try:
                raise validation_result
            except ValidationWarning as vw:
                warnings.warn(vw.message)

    def _collect_validation_results(self) -> list[ValidationResult]:
        sensor_config_validate_results = (
            self._validate_continuous_sweep_mode()
            + self._validate_idle_states()
            + self._validate_sweep_and_frame_rate()
            + self._validate_required_buffer_usage()
            + self._validate_ascending_prf_in_subsweeps()
        )
        subsweep_config_validate_results = []
        for subsweep in self.subsweeps:
            subsweep_config_validate_results.extend(subsweep._collect_validation_results())
        return sensor_config_validate_results + subsweep_config_validate_results

    def _validate_continuous_sweep_mode(self) -> list[ValidationResult]:
        if not self.continuous_sweep_mode:
            return []

        validation_results: list[ValidationResult] = []

        if self.frame_rate is not None:
            validation_results.append(
                ValidationError(
                    self,
                    "frame_rate",
                    "Frame rate must be unset to use continuous sweep mode.",
                )
            )
        if self.sweep_rate is None:
            validation_results.append(
                ValidationError(
                    self, "sweep_rate", "Sweep rate must be set to use continuous sweep mode."
                )
            )
        if self.inter_frame_idle_state != self.inter_sweep_idle_state:
            validation_results.extend(
                ValidationError(
                    self,
                    idle_state,
                    "Inter sweep/frame idle states must be equal to use continuous sweep mode.",
                )
                for idle_state in ["inter_frame_idle_state", "inter_sweep_idle_state"]
            )
        if self.sweeps_per_frame == 1:
            validation_results.append(
                ValidationWarning(
                    self,
                    "continuous_sweep_mode",
                    (
                        "Not meaningful with only 1 sweep per frame. A fixed frame rate can be "
                        "used instead."
                    ),
                )
            )
        return validation_results

    def _validate_idle_states(self) -> list[ValidationResult]:
        if not (
            self.inter_frame_idle_state.is_deeper_than(self.inter_sweep_idle_state)
            or self.inter_frame_idle_state == self.inter_sweep_idle_state
        ):
            return [
                ValidationError(
                    self,
                    idle_state,
                    "Inter frame idle state needs to be deeper "
                    + "or the same as inter sweep idle state",
                )
                for idle_state in ["inter_frame_idle_state", "inter_sweep_idle_state"]
            ]
        return []

    def _validate_sweep_and_frame_rate(self) -> list[ValidationResult]:
        validation_results: list[ValidationResult] = []

        if self.sweep_rate is not None and self.frame_rate is not None:
            seconds_needed_per_frame = self.sweeps_per_frame / self.sweep_rate

            if self.frame_rate > 1 / seconds_needed_per_frame:
                validation_results.append(
                    ValidationError(
                        self,
                        "frame_rate",
                        "The frame rate is set higher than what the sweep rate allows.\n"
                        + f"Frame rate: {self.frame_rate:.2f} Hz\n"
                        + "Sweep rate / sweeps per frame: "
                        + f"{self.sweep_rate / self.sweeps_per_frame:.2f} Hz",
                    )
                )

            if np.isclose(self.frame_rate, 1 / seconds_needed_per_frame):
                validation_results.extend(
                    ValidationWarning(
                        self,
                        field,
                        "Frame rate is approximately equal to SPF / Sweep rate. "
                        + "Use continuous sweep mode instead.",
                    )
                    for field in ["sweep_rate", "frame_rate", "sweeps_per_frame"]
                )
        return validation_results

    def _validate_required_buffer_usage(self) -> list[ValidationResult]:
        BUFFER_SIZE = 4096
        ERROR_MSG = "Required buffer size is too large: {}/{}."

        buffer_size_available = (
            (BUFFER_SIZE // 2) - 1 if self.double_buffering else BUFFER_SIZE - 1
        )
        total_num_points = sum(subsweep_config.num_points for subsweep_config in self.subsweeps)
        required_buffer_size = total_num_points * self.sweeps_per_frame

        validation_results: list[ValidationResult] = []

        if required_buffer_size > buffer_size_available:
            validation_results.append(
                ValidationError(
                    source=self,
                    aspect="sweeps_per_frame",
                    message=(
                        ERROR_MSG.format(required_buffer_size, buffer_size_available)
                        + " Decreasing sweeps per frame reduces buffer usage."
                    ),
                    tag=VALIDATION_TAG_BUFFER_SIZE_TOO_LARGE,
                ),
            )
            validation_results.extend(
                ValidationError(
                    source=subsweep_config,
                    aspect="num_points",
                    message=(
                        ERROR_MSG.format(required_buffer_size, buffer_size_available)
                        + " Decreasing number of points reduces buffer usage."
                    ),
                    tag=VALIDATION_TAG_BUFFER_SIZE_TOO_LARGE,
                )
                for subsweep_config in self.subsweeps
            )
        return validation_results

    def _validate_ascending_prf_in_subsweeps(self) -> list[ValidationResult]:
        """Validate that PRFs are defined in an ascending fashion with subsweeps.
        I.e. Subsweep index and PRF should be inversly proportional.
        """

        if len(self.subsweeps) == 1:
            return []

        validation_results: list[ValidationResult] = []

        for (ss_curr_idx, ss_curr), (ss_next_idx, ss_next) in zip(
            enumerate(self.subsweeps),
            enumerate(self.subsweeps[1:], start=1),
        ):
            if ss_curr.prf.frequency < ss_next.prf.frequency:
                ss_curr_valid_prfs_names = ", ".join(
                    prf.name for prf in PRF if (prf.frequency >= ss_next.prf.frequency)
                )
                validation_results.append(
                    ValidationError(
                        source=ss_curr,
                        aspect="prf",
                        message=f"subsweeps[{ss_curr_idx}].prf = {ss_curr.prf.name}. It needs to be one of [{ss_curr_valid_prfs_names}] because subsweeps[{ss_next_idx}].prf = {ss_next.prf.name}",
                    )
                )
                ss_next_valid_prfs_names = ", ".join(
                    prf.name for prf in PRF if (prf.frequency <= ss_curr.prf.frequency)
                )
                validation_results.append(
                    ValidationError(
                        source=ss_next,
                        aspect="prf",
                        message=f"subsweeps[{ss_next_idx}].prf = {ss_next.prf.name}. It needs to be one of [{ss_next_valid_prfs_names}] because subsweeps[{ss_curr_idx}].prf = {ss_curr.prf.name}",
                    )
                )
        return validation_results

    @property
    def sweeps_per_frame(self) -> int:
        """Sweeps per frame (SPF)

        The number of sweeps that will be captured in each frame (measurement).

        Must be > 0.
        """

        return self._sweeps_per_frame

    @sweeps_per_frame.setter
    def sweeps_per_frame(self, value: int) -> None:
        int_value = utils.convert_validate_int(value, min_value=1)
        self._sweeps_per_frame = int_value

    @property
    def sweep_rate(self) -> Optional[float]:
        """Sweep rate

        The sweep rate for sweeps in a frame (measurement).

        In Hz. Must be > 0 or ``None``, where ``None`` is interpreted as max sweep rate.
        """

        return self._sweep_rate

    @sweep_rate.setter
    def sweep_rate(self, value: Optional[float]) -> None:
        if value is None:
            self._sweep_rate = None
        else:
            self._sweep_rate = utils.validate_float(value, min_value=0.0, inclusive=False)

    @property
    def frame_rate(self) -> Optional[float]:
        """Frame rate

        Setting the frame rate to unlimited means that the rate is not limited by the sensor but
        the rate that the host acknowledge and reads out the measurement data.

        In Hz. Must be > 0 or ``None``, where ``None`` is interpreted as unlimited.
        """

        return self._frame_rate

    @frame_rate.setter
    def frame_rate(self, value: Optional[float]) -> None:
        if value is None:
            self._frame_rate = None
        else:
            self._frame_rate = utils.validate_float(value, min_value=0.0, inclusive=False)

    @property
    def continuous_sweep_mode(self) -> bool:
        """Continuous sweep mode (CSM)

        With CSM, the sensor timing is set up to generate a continuous
        stream of sweeps, even if more than one sweep per frame is used.
        The interval between the last sweep in one frame to the first
        sweep in the next frame becomes equal to the interval between
        sweeps within a frame (given by the sweep rate).

        It ensures that:

        'frame rate' = 'sweep rate' / 'sweeps per frame'

        While the frame rate parameter can be set to approximately
        satisfy this condition, using CSM is more precise.

        If only one sweep per frame is used, CSM has no use since a
        continuous stream of sweeps is already given (if a fixed frame
        rate is used).

        The main use for CSM is to allow reading out data at a slower
        rate than the sweep rate, while maintaining that sweep rate
        continuously.

        Note that in most cases, double buffering must be enabled to
        allow high rates without delays.

        Constraints:

        - :attr:`frame_rate` must be set to unlimited (``None``).
        - :attr:`sweep_rate` must be set (> 0).
        - :attr:`inter_frame_idle_state` must be set equal to :attr:`inter_sweep_idle_state`.
        """

        return self._continuous_sweep_mode

    @continuous_sweep_mode.setter
    def continuous_sweep_mode(self, value: bool) -> None:
        self._continuous_sweep_mode = bool(value)

    @property
    def double_buffering(self) -> bool:
        """Double buffering

        If enabled, the sensor buffer will be split in two halves reducing the
        maximum number of samples. A frame can be read while sampling is done into the
        other buffer.

        When using double buffering, measurements coinciding with SPI activity may have distorted
        phase. To mitigate this issue, applying a median filter is recommended.
        """

        return self._double_buffering

    @double_buffering.setter
    def double_buffering(self, value: bool) -> None:
        self._double_buffering = bool(value)

    @property
    def inter_frame_idle_state(self) -> IdleState:
        """Inter frame idle state

        The inter frame idle state is the state the sensor idles in between each frame.

        Idle state `Deep sleep` is the deepest state where as much of the
        sensor hardware as possible is shut down and idle state `Ready` is
        the lightest state where most of the sensor hardware is kept on.

        `Deep sleep` is the slowest to transition from while `Ready` is the fastest.

        The :attr:`inter_frame_idle_state` of the frame must be deeper or
        the same as the :attr:`inter_sweep_idle_state` .
        """

        return self._inter_frame_idle_state

    @inter_frame_idle_state.setter
    def inter_frame_idle_state(self, value: IdleState) -> None:
        self._inter_frame_idle_state = IdleState(value)

    @property
    def inter_sweep_idle_state(self) -> IdleState:
        """Inter sweep idle state

        The inter sweep idle state is the state the sensor idles in
        between each sweep in a frame.

        Idle state `Deep sleep` is the deepest state where as much of the
        sensor hardware as possible is shut down and idle state `Ready` is
        the lightest state where most of the sensor hardware is kept on.

        `Deep sleep` is the slowest to transition from while `Ready` is the fastest.
        """

        return self._inter_sweep_idle_state

    @inter_sweep_idle_state.setter
    def inter_sweep_idle_state(self, value: IdleState) -> None:
        self._inter_sweep_idle_state = IdleState(value)

    def _pretty_str_lines(self, sensor_id: Optional[int] = None) -> list[str]:
        lines = []

        id_str = "" if sensor_id is None else f" @ sensor {sensor_id}"
        lines.append(f"{type(self).__name__}{id_str}:")

        d = self.to_dict()
        del d["subsweeps"]
        lines.extend(pretty_dict_line_strs(d))

        lines.append("  subsweeps:")
        for i, subsweep in enumerate(self.subsweeps):
            ss_lines = subsweep._pretty_str_lines(index=i)
            lines.extend(utils.indent_strs(ss_lines, 2))

        return lines

    def __str__(self) -> str:
        return "\n".join(self._pretty_str_lines())

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            + f"subsweeps={self.subsweeps}, "
            + f"sweeps_per_frame={self.sweeps_per_frame}, "
            + f"sweep_rate={self.sweep_rate}, "
            + f"frame_rate={self.frame_rate}, "
            + f"continuous_sweep_mode={self.continuous_sweep_mode}, "
            + f"double_buffering={self.double_buffering}, "
            + f"inter_frame_idle_state={self.inter_frame_idle_state}, "
            + f"inter_sweep_idle_state={self.inter_sweep_idle_state}"
            + ")"
        )


opser.register_json_presentable(SensorConfig)



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import typing as t

from PySide6.QtWidgets import QTextBrowser, QWidget


_HELP_MD = """
# Acconeer Exploration Tool Help

Welcome to the help section of Exploration Tool.
If you are searching for information, further reading or software downloads,
these links might be of help:

## Information and further reading

- [Handbook](https://docs.acconeer.com/en/latest/handbook/index.html):
Overview of radar concepts, information about Acconeer's radars.

- [Exploration Tool docs](https://docs.acconeer.com/en/latest/exploration_tool/index.html):
Covers how to install Exploration Tool and EVK setup,
Python API reference and algorithm descriptions.

- [docs.acconeer.com](https://docs.acconeer.com):
Documentation portal

## Software downloads

- [Acconeer on Github](https://github.com/acconeer):
Hosts machine learning examples and Exploration Tool itself.<br />Machine learning examples:
    + [Gesture control](https://github.com/acconeer/acconeer-a121-gesture-control)
    + [Grass detection](https://github.com/acconeer/acconeer-a121-grass-detection)
    + [Carpet detection](https://github.com/acconeer/acconeer-a121-carpet-detection)

- [Developer Site](https://developer.acconeer.com):
User guides and C SDKs.

## Other links:

- [Acconeer's Products](https://www.acconeer.com/products/)

- [Acconeer on Twitter](https://twitter.com/acconeer_ab)

- [Acconeer on Instagram](https://www.instagram.com/acconeerab/)

"""


class HelpMainWidget(QTextBrowser):
    def __init__(self, parent: t.Optional[QWidget] = None) -> None:
        super().__init__()

        self.setReadOnly(True)
        self.setOpenExternalLinks(True)

        self.setStyleSheet("border: none;")
        self.setMarkdown(_HELP_MD)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import typing as t

import typing_extensions as te

from acconeer.exptool._core.entities import ClientInfo


_ConfigT = t.TypeVar("_ConfigT", contravariant=True)
_MetadataT = t.TypeVar("_MetadataT", contravariant=True)
_ResultT = t.TypeVar("_ResultT", contravariant=True)
_ServerInfoT = t.TypeVar("_ServerInfoT", contravariant=True)

_RecorderT = t.TypeVar("_RecorderT")


class RecorderAttachable(te.Protocol[_RecorderT]):
    """Dependecy Inversion interface for clients"""

    def attach_recorder(self, recorder: _RecorderT) -> None: ...

    def detach_recorder(self) -> t.Optional[_RecorderT]: ...


class Recorder(te.Protocol[_ConfigT, _MetadataT, _ResultT, _ServerInfoT]):
    """The interface a Recorder needs to follow"""

    def _start(
        self,
        *,
        client_info: ClientInfo,
        server_info: _ServerInfoT,
    ) -> None: ...

    def _start_session(self, *, config: _ConfigT, metadata: _MetadataT) -> None: ...

    def _sample(self, result: _ResultT) -> None: ...

    def _stop_session(self) -> None: ...

    def close(self) -> t.Any: ...

    def __enter__(self) -> te.Self:
        return self

    def __exit__(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.close()



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
from __future__ import annotations

from typing import Any, Dict, Tuple

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    exponential_smoothing_coefficient,
    get_distances_m,
    get_temperature_adjustment_factors,
)


MAX_AMPLITUDE = 46341  # sqrt((2^15)^2 + (2^15)^2)


@attrs.mutable(kw_only=True)
class ObstructionProcessorConfig:
    distance_threshold: float = attrs.field(default=0.1)
    """How far the weighted phase signature can be from calibration value and still be considered unobstructed."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []
        return validation_results


@attrs.frozen(kw_only=True)
class ObstructionProcessorExtraResult:
    obstruction_signature: npt.NDArray[np.float64] = attrs.field()
    """Signature of amplitudes used for the obstruction detection."""

    obstruction_data: npt.NDArray[np.float64] = attrs.field()
    """Array with obstruction amplitudes (usually direct leakage)."""

    obstruction_center: npt.NDArray[np.float64] = attrs.field()
    """Signature of amplitudes used in calibration."""

    obstruction_distance: float = attrs.field()
    """Distance threshold, same as in configuration."""


@attrs.mutable(kw_only=True)
class ObstructionProcessorResult:
    obstruction_found: bool = attrs.field(default=False)
    """Boolean indicating whether something is obstructing the sensor."""

    extra_result: ObstructionProcessorExtraResult = attrs.field(default=None)
    """Extra information for plotting."""


class ObstructionProcessor:
    # Cannot be ProcessorBase since the processor takes a frame instead of result
    def __init__(
        self,
        sensor_config: a121.SensorConfig,
        processor_config: ObstructionProcessorConfig,
        metadata: a121.Metadata,
        update_rate: float,
        calibration_center: npt.NDArray[np.float64],
        calibration_noise_mean: float,
        calibration_temperature: float,
    ):
        self.profile = sensor_config.profile
        self.processor_config = processor_config

        self.dist_thres = self.processor_config.distance_threshold

        self.distances = get_distances_m(sensor_config, metadata)

        self.x_dist_threshold, self.y_dist_threshold = self.get_thresholds(
            self.dist_thres, self.distances
        )

        self.calibration_noise_mean = calibration_noise_mean
        self.calibration_center = calibration_center
        self.calibration_temperature = calibration_temperature

        self.lp_signature = np.array(calibration_center)
        self.lp_const = exponential_smoothing_coefficient(
            update_rate, 0.8
        )  # 0.8 seconds to deal with noise jitter

    @staticmethod
    def get_thresholds(
        distance_threshold: float, distances: npt.NDArray[np.float64]
    ) -> Tuple[float, float]:
        distance_range = max(distances) - min(distances)

        x_width = distance_range * distance_threshold
        y_width = MAX_AMPLITUDE * distance_threshold

        return x_width, y_width

    @staticmethod
    def get_signature(
        frame: npt.NDArray[np.complex128], noise_level: float, distances: npt.NDArray[np.float64]
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        # Returns signature for a frame without temperature adjustment

        abs_frame = np.squeeze(abs(frame))
        noise_adjusted_frame = abs_frame - noise_level
        total_energy = np.sum(noise_adjusted_frame)
        avg_energy = total_energy / len(distances)
        weighted_sum = np.sum(distances * noise_adjusted_frame) / total_energy
        signature = np.array([weighted_sum, avg_energy])
        return signature, noise_adjusted_frame

    @staticmethod
    def get_noise_level(frame: npt.NDArray[np.complex128]) -> float:
        mean_frame = np.mean(frame, axis=0)
        abs_frame = abs(mean_frame)
        noise_level = float(np.mean(abs_frame))
        return noise_level

    def process(
        self, frame: npt.NDArray[np.complex128], temperature: float
    ) -> ObstructionProcessorResult:
        # Do temperature adjustment
        signal_adjustment_factor, deviation_adjustment_factor = get_temperature_adjustment_factors(
            reference_temperature=self.calibration_temperature,
            current_temperature=temperature,
            profile=self.profile,
        )

        # We could recalculate the level of the calibration frame as well.
        # That would be done by calibration_frame * signal_adjustment_factor
        temp_adjusted_frame = frame / signal_adjustment_factor

        noise_level = self.calibration_noise_mean * deviation_adjustment_factor

        signature, amp_adjusted = self.get_signature(
            temp_adjusted_frame, noise_level, self.distances
        )

        # low pass filtering to amend the noise jitter
        self.lp_signature = self.lp_signature * self.lp_const + (1.0 - self.lp_const) * signature

        dist_x = abs(self.lp_signature[0] - self.calibration_center[0])
        dist_y = abs(self.lp_signature[1] - self.calibration_center[1])

        obstructed = (dist_y > self.y_dist_threshold) or (dist_x > self.x_dist_threshold)

        extra_result = ObstructionProcessorExtraResult(
            obstruction_signature=self.lp_signature,
            obstruction_data=amp_adjusted,
            obstruction_center=self.calibration_center,
            obstruction_distance=self.dist_thres,
        )
        res = ObstructionProcessorResult(obstruction_found=obstructed, extra_result=extra_result)
        return res


@attrs.mutable(kw_only=True)
class ProcessorConfig:
    queue_length: int = attrs.field(default=5)
    """Number of samples above threshold to trigger parked car."""

    amplitude_threshold: float = attrs.field(default=15.0)
    """Threshold level in times the noise level to trigger."""

    weighted_distance_threshold_m: float = attrs.field(default=0.1)
    """Threshold in m between weighted distance points to consider them similar."""

    signature_similarity_threshold: float = attrs.field(default=0.6)
    """How large fraction of the signature history that has to be similar in order to retain detection."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []
        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    signature_history: npt.NDArray[np.float64] = attrs.field()
    """Array containing queue_length last signatures."""

    parking_data: npt.NDArray[np.float64] = attrs.field()
    """The scaled amplitude array used to calculate the last signature."""

    closest_observation: float = attrs.field()
    """If a car is detected, the distance to the weighted closest observation."""


@attrs.mutable(kw_only=True)
class ProcessorResult:
    car_detected: bool = attrs.field(default=False)
    """If a car (or other large object) is detected in front of the sensor."""

    extra_result: ProcessorExtraResult = attrs.field()
    """Extra information for plotting."""


class Processor:
    # Cannot be ProcessorBase since the processor takes a frame instead of result
    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        metadata: a121.Metadata,
        noise_estimate: float = 1.0,
        noise_estimate_temperature: float = 0.0,
    ):
        self.profile = sensor_config.profile
        self.distances = get_distances_m(sensor_config, metadata)
        if noise_estimate == 0.0:
            noise_estimate = 1.0
        self.noise_estimate = noise_estimate
        self.noise_estimate_temperature = noise_estimate_temperature

        self.norm_distances = self.distances / self.distances[0]

        self.frame_ind = 0

        # Thresholds
        self.weight_threshold = processor_config.amplitude_threshold
        self.distance_threshold = processor_config.weighted_distance_threshold_m
        self.similarity_threshold = processor_config.signature_similarity_threshold

        # signature history
        self.queue_length = processor_config.queue_length
        self.sig_history = np.zeros(
            self.queue_length,
            dtype=[("weighted_distance", float), ("max_energy", float)],
        )

    @classmethod
    def process_noise_frame(cls, frame: npt.NDArray[np.complex128]) -> float:
        n_std_dev = 2
        mean_frame = np.mean(frame, axis=0)
        amp = np.abs(mean_frame)
        dev = np.std(amp)
        noise_level = float(np.mean(amp) + n_std_dev * dev)
        return noise_level

    def signature(self, depths: npt.NDArray[np.float64]) -> Tuple[float, float]:
        total_energy = sum(depths)
        max_energy = max(depths)
        if total_energy != 0.0:
            weighted_distance = sum(depths * self.distances) / total_energy
        else:
            weighted_distance = self.distances[
                0
            ]  # We should return a distance, so pick the first.
        return (weighted_distance, max_energy)

    def objects_present(self) -> bool:
        energy_history = np.array([elm[1] for elm in self.sig_history])
        n_trigs = sum(energy_history > self.weight_threshold)
        ret = n_trigs > (self.queue_length * self.similarity_threshold)
        return ret

    def same_objects(self) -> Dict[str, Any]:
        depth_sigs = np.sort(self.sig_history, axis=0, order=["weighted_distance", "max_energy"])
        weights = np.array([elm[1] for elm in depth_sigs])
        depth_sigs = depth_sigs[weights > self.weight_threshold]

        clusters = []
        curr_cluster: list[Tuple[float, float]] = []
        for elm in depth_sigs:
            if len(curr_cluster) == 0:
                curr_cluster.append(elm)
            else:
                if (elm[0] - curr_cluster[0][0]) > self.distance_threshold:
                    clusters.append(curr_cluster)
                    curr_cluster = []
                else:
                    curr_cluster.append(elm)
        if len(curr_cluster) > 0:
            clusters.append(curr_cluster)

        if len(clusters) > 0:
            cluster_lengths = np.array([len(cluster) for cluster in clusters])
            long_cluster_ind = np.argmax(cluster_lengths)
            closest_dist = clusters[long_cluster_ind][0][0]
            similarity = max(cluster_lengths) / self.queue_length
            detection = similarity > self.similarity_threshold
        else:
            detection = False
            closest_dist = 0

        ret = {"detection": detection, "closest_dist": closest_dist}

        return ret

    def process(self, frame: npt.NDArray[np.complex128], temperature: float) -> ProcessorResult:
        self.frame_ind += 1

        _, deviation_adjustment_factor = get_temperature_adjustment_factors(
            reference_temperature=self.noise_estimate_temperature,
            current_temperature=temperature,
            profile=self.profile,
        )

        noise_level_adjusted = self.noise_estimate * deviation_adjustment_factor

        data = np.squeeze(frame, axis=0)  # remove sweep dimension

        # Div by zero avoided by tests elsewhere
        amp = np.squeeze(abs(data)) / noise_level_adjusted

        amp_noise_removed = np.fmax(amp - 1.0, 0)
        amp_scaled = amp_noise_removed * self.norm_distances

        sig = self.signature(amp_scaled)

        self.sig_history = np.roll(self.sig_history, -1, axis=0)
        self.sig_history[-1] = sig

        objects_present = self.objects_present()
        same_objects_info = self.same_objects()
        same_objects = same_objects_info["detection"]
        closest_object = same_objects_info["closest_dist"]

        parked_car = objects_present and same_objects

        extra_result = ProcessorExtraResult(
            signature_history=self.sig_history,
            parking_data=amp_scaled,
            closest_observation=closest_object,
        )

        ret = ProcessorResult(car_detected=parked_car, extra_result=extra_result)
        return ret



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

from abc import ABC, abstractclassmethod, abstractmethod
from pathlib import Path
from typing import Optional, Sequence, Tuple, Union


try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

import acconeer.exptool as et


_Description = str
_FileExtension = str
_Path = Union[str, Path]
_AcceptedFileExtensions = Sequence[Tuple[_FileExtension, _Description]]


class Calibration(ABC):
    @abstractmethod
    def save(self, path: _Path): ...

    @abstractclassmethod
    def load(cls, path: _Path) -> Calibration: ...

    @abstractclassmethod
    def file_extensions(cls) -> _AcceptedFileExtensions: ...

    @classmethod
    def validate_path(
        cls, path: _Path, file_extensions: Optional[_AcceptedFileExtensions] = None
    ) -> None:
        """
        :param path: Path to validate
        :param file_extensions: Override the class-defined file extensions to accept.

        :raises: ValueError if path is not an acceptable path.
        """
        file_extensions = file_extensions or cls.file_extensions()

        extension = Path(path).suffix.strip(".")
        valid_extensions = [ext.strip(".") for ext, _ in file_extensions]

        extension_ok = extension in valid_extensions

        if not extension_ok:
            raise ValueError(
                (
                    f'Extension of file "{path}" ("{extension}") is not valid.'
                    + f"Should be one of {valid_extensions}."
                )
            )


class CalibrationMapper(Protocol):
    """
    This mapper is needed to allow for Configurations to contain a subset of the parameters
    in a Calibration. The configurable parameters (members of CalibrationConfiguration) are
    a subset (strict or non-strict) of the "actual" parameters in the Calibration.
    """

    @abstractclassmethod
    def get_updated_calibration_from_configuration(
        cls, configuration: et.configbase.Config, calibration: Optional[Calibration]
    ) -> Calibration:
        """
        Creates a new Calibration instance given a Config and (optionally) a Calibration.
        Fields in the Config should have precedence over the fields in the passed Calibration.

        This function is allowed to raise a ValueError if calibration=None is not supported.
        """
        ...

    @abstractclassmethod
    def update_config_from_calibration(
        cls, configuration: et.configbase.Config, calibration: Calibration
    ) -> None:
        """
        Updates the passed Config with the fields of the passed calibration.
        """
        ...



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from enum import Enum, auto


class ConnectionState(Enum):
    DISCONNECTED = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    DISCONNECTING = auto()


class ConnectionInterface(Enum):
    SERIAL = auto()
    SOCKET = auto()
    USB = auto()
    SIMULATED = auto()


class PluginState(Enum):
    """Describes what state the plugin is in.

    Theese are the possible state transitions:

           +-------+      +---------------+
        +->|LOADING|-+  +>|LOADED_STARTING|-+
        |  +-------+ |  | +---------------+ |
        |            v  |                   v
    +---+----+   +------+----+      +-----------+
    |UNLOADED|   |LOADED_IDLE|      |LOADED_BUSY|
    +--------+   +----+------+      +-------+---+
        ^             | ^                   |
        | +---------+ | | +---------------+ |
        +-|UNLOADING|<+ +-|LOADED_STOPPING|<+
          +---------+     +---------------+
    """

    UNLOADED = auto()
    UNLOADING = auto()
    LOADING = auto()
    LOADED_IDLE = auto()
    LOADED_STARTING = auto()
    LOADED_BUSY = auto()
    LOADED_STOPPING = auto()

    @property
    def is_loaded(self) -> bool:
        return self in {
            self.LOADED_IDLE,
            self.LOADED_STARTING,
            self.LOADED_BUSY,
            self.LOADED_STOPPING,
        }

    @property
    def is_busy(self) -> bool:
        return self in {
            self.LOADED_STARTING,
            self.LOADED_BUSY,
            self.LOADED_STOPPING,
        }

    @property
    def is_steady(self) -> bool:
        return self in {
            self.UNLOADED,
            self.LOADED_IDLE,
        }


class PluginFamily(Enum):
    SERVICE = "Services"
    DETECTOR = "Detectors"
    REF_APP = "Reference apps"
    EXAMPLE_APP = "Example apps"
    EXTERNAL_PLUGIN = "External Plugins"


class PluginGeneration(Enum):
    A111 = "a111"
    A121 = "a121"



# Copyright (c) Acconeer AB, 2024
# All rights reserved

from __future__ import annotations

import numpy as np

import acconeer.exptool.a121 as a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    select_prf,
)
from acconeer.exptool.a121.algo._utils import (
    get_max_profile_without_direct_leakage,
    get_max_step_length,
)


MAX_NUM_SUBSWEEPS = 4
FWHM_MARGIN_FACTOR = 2.0
MIN_DIST_M = {
    a121.Profile.PROFILE_1: 0,
    a121.Profile.PROFILE_2: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_2],
    a121.Profile.PROFILE_3: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_3],
    a121.Profile.PROFILE_4: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_4],
    a121.Profile.PROFILE_5: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_5],
}
BREAK_POINTS_MAP = {
    1: (0, 1.0),
    2: (0, 0.74, 1.0),
    3: (0, 0.54, 0.81, 1.0),
    4: (0, 0.47, 0.70, 0.87, 1.0),
}

"""
Functions to select hwaas and profile for different subsweeps.
"""


def _get_hwaas(signal_quality: float, subsweep_end_point_m: float, profile: a121.Profile) -> int:
    RLG_PER_HWAAS_MAP = {
        a121.Profile.PROFILE_1: 11.3,
        a121.Profile.PROFILE_2: 13.7,
        a121.Profile.PROFILE_3: 19.0,
        a121.Profile.PROFILE_4: 20.5,
        a121.Profile.PROFILE_5: 21.6,
    }

    rlg = signal_quality + 4 * 10 * np.log10(subsweep_end_point_m)
    hwaas_in_subsweep = int(round(10 ** ((rlg - RLG_PER_HWAAS_MAP[profile]) / 10)))
    hwaas_in_subsweep = np.clip(hwaas_in_subsweep, 1, 511)

    return hwaas_in_subsweep


def _get_profile_conf(
    min_profile: a121.Profile, max_profile: a121.Profile, n_subsweeps: int
) -> list[a121.Profile]:
    """
    Function to select profiles for subsweeps.
    The function will eliminate profiles according to the "remove_order".
    """
    base = [
        a121.Profile.PROFILE_1,
        a121.Profile.PROFILE_2,
        a121.Profile.PROFILE_3,
        a121.Profile.PROFILE_4,
    ]
    base = base[base.index(min_profile) : base.index(max_profile) + 1]
    remove_order = [a121.Profile.PROFILE_2, a121.Profile.PROFILE_4, a121.Profile.PROFILE_3]
    for elm in remove_order:
        if len(base) <= n_subsweeps:
            return base
        if elm != base[0] and elm in base:
            base.remove(elm)
    return base


def _get_profile_subsweeps(
    start_m: float,
    end_m: float,
    max_n_subsweeps: int,
    signal_quality: float,
    add_final_point: bool,
) -> list[a121.SubsweepConfig]:
    """
    Splits up the subsweeps with different profiles.
    This function only deals with Profile 4 and lower.
    The hwaas will also differ, but is not the driver.
    Adds all points not including end_m (unless start == end, then add one point).
    If add_final_point is true adds an extra point (possibly after end_m).
    """

    if start_m == end_m:
        add_final_point = False
        end_m += 0.01

    min_profile = get_max_profile_without_direct_leakage(start_m)
    max_profile = get_max_profile_without_direct_leakage(end_m)

    if min_profile == a121.Profile.PROFILE_5:
        min_profile = a121.Profile.PROFILE_4
    if max_profile == a121.Profile.PROFILE_5:
        max_profile = a121.Profile.PROFILE_4
    profiles_to_use = _get_profile_conf(min_profile, max_profile, max_n_subsweeps)

    curr_point = start_m
    subsweeps_configs = []

    cut_points = [MIN_DIST_M[profile] for profile in profiles_to_use] + [end_m]

    for profile, cut_point in zip(profiles_to_use, cut_points[1:]):
        curr_start = curr_point
        n_points_in_subsweep = 0
        step = get_max_step_length(profile)
        while curr_point < cut_point:
            curr_point += step * APPROX_BASE_STEP_LENGTH_M
            n_points_in_subsweep += 1
        curr_point_ind = int(round(curr_point / APPROX_BASE_STEP_LENGTH_M))
        subsweeps_configs.append(
            a121.SubsweepConfig(
                profile=profile,
                start_point=int(round(curr_start / APPROX_BASE_STEP_LENGTH_M)),
                step_length=step,
                num_points=n_points_in_subsweep,
                hwaas=_get_hwaas(signal_quality, curr_point, profile),
                prf=select_prf(curr_point_ind, profile),
            )
        )
    if add_final_point:
        subsweeps_configs[-1].num_points += 1
    return subsweeps_configs


def _get_hwaas_subsweeps(
    start_m: float,
    end_m: float,
    max_n_subsweeps: int,
    signal_quality: float,
    add_final_point: bool,
) -> list[a121.SubsweepConfig]:
    """
    Split up the subsweeps with different hwaas (further -> higher hwaas).
    Assumes that all subsweeps uses profile 5.
    Adds all points not including end_m (unless start == end, then add one point).
    If add_final_point is true adds an extra point (possibly after end_m).
    """

    if start_m == end_m:
        add_final_point = False
        end_m += 0.01

    profile = a121.Profile.PROFILE_5
    step = get_max_step_length(profile)
    break_points_fracs = BREAK_POINTS_MAP[max_n_subsweeps]
    cut_points = start_m + np.array(break_points_fracs) * (end_m - start_m)
    curr_point = start_m
    subsweeps_configs = []
    for cut_point in cut_points[1:]:
        n_points_in_subsweep = 0
        curr_start = curr_point
        while curr_point < cut_point:
            curr_point += step * APPROX_BASE_STEP_LENGTH_M
            n_points_in_subsweep += 1
        if n_points_in_subsweep > 0:
            curr_point_ind = int(round(curr_point / APPROX_BASE_STEP_LENGTH_M))
            subsweeps_configs.append(
                a121.SubsweepConfig(
                    profile=profile,
                    start_point=int(round(curr_start / APPROX_BASE_STEP_LENGTH_M)),
                    step_length=step,
                    num_points=n_points_in_subsweep,
                    hwaas=_get_hwaas(signal_quality, curr_point, profile),
                    prf=select_prf(curr_point_ind, profile),
                )
            )
    if add_final_point:
        subsweeps_configs[-1].num_points += 1
    return subsweeps_configs


def get_subsweep_configs(
    start_m: float, end_m: float, signal_quality: float
) -> list[a121.SubsweepConfig]:
    """
    The function separates on two cases, closer than profile 5 and further than profile 5.
    The main driver for the close case will be to use different profiles for the subsweeps.
    For the far case, all subsweeps will use profile 5 and instead have different hwaas.
    If both cases is applicable, a split will occur.
    For each 64 cm another subsweep will be added to the "far" case up to three subsweeps.
    """

    # clip to closest multiple of APPROX_BASE_STEP_LENGTH_M
    start_m = np.ceil(start_m / APPROX_BASE_STEP_LENGTH_M) * APPROX_BASE_STEP_LENGTH_M

    end_m = np.ceil(end_m / APPROX_BASE_STEP_LENGTH_M) * APPROX_BASE_STEP_LENGTH_M

    profile_5_breakpoint = MIN_DIST_M[a121.Profile.PROFILE_5]

    if end_m < profile_5_breakpoint:
        ss = _get_profile_subsweeps(start_m, end_m, 4, signal_quality, add_final_point=True)
    elif start_m >= profile_5_breakpoint:
        ss = _get_hwaas_subsweeps(start_m, end_m, 4, signal_quality, add_final_point=True)
    else:
        n_hwaas_subs = max(min(3, int(round(np.ceil(end_m / profile_5_breakpoint))) - 1), 1)
        n_profile_subs = 4 - n_hwaas_subs
        ss_1 = _get_profile_subsweeps(
            start_m, profile_5_breakpoint, n_profile_subs, signal_quality, add_final_point=False
        )
        ss_2 = _get_hwaas_subsweeps(
            profile_5_breakpoint, end_m, n_hwaas_subs, signal_quality, add_final_point=True
        )
        ss = ss_1 + ss_2
    return ss



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from . import algo
from .api import (
    configured_rate,
    converged_average_current,
    frame_active,
    frame_idle,
    group_active,
    group_idle,
    power_state,
    session,
    subsweep_active,
    sweep_active,
    sweep_idle,
)
from .domain import (
    CompositeRegion,
    EnergyRegion,
    SimpleRegion,
)
from .lookup import (
    Module,
    Sensor,
)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
from __future__ import annotations

import logging
import multiprocessing as mp
import platform
import queue
import signal
import threading as th
import traceback
from time import sleep, time
from typing import Optional, Tuple, Union

import serial

from .buffered_link import BufferedLink, LinkError


log = logging.getLogger(__name__)


class BaseSerialLink(BufferedLink):
    def __init__(self, baudrate: int = 115200) -> None:
        super().__init__()
        self._timeout = self.DEFAULT_TIMEOUT
        self._baudrate = baudrate


class SerialLink(BaseSerialLink):
    def __init__(self, port: Optional[str] = None, flowcontrol: bool = False) -> None:
        super().__init__()
        self._port = port
        self._ser: Optional[serial.Serial] = None
        self._flowcontrol = flowcontrol

    def _update_timeout(self) -> None:
        if self._ser is not None:
            self._ser.timeout = self._timeout

    def send_break(self) -> None:
        assert self._ser is not None
        self._ser.send_break()
        sleep(1.0)
        self._ser.flushInput()

    def connect(self) -> None:
        self._ser = serial.Serial()
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._update_timeout()
        self._ser.open()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

    def recv(self, num_bytes: int) -> bytes:
        assert self._ser is not None
        data = bytearray(self._ser.read(num_bytes))

        if len(data) != num_bytes:
            msg = "recv timeout"
            raise LinkError(msg)

        return data

    def recv_until(self, bs: bytes) -> bytes:
        assert self._ser is not None
        data = bytearray(self._ser.read_until(bs))

        if bs not in data:
            msg = "recv timeout"
            raise LinkError(msg)

        return data

    def send(self, data: bytes) -> None:
        assert self._ser is not None
        self._ser.write(data)

    def disconnect(self) -> None:
        if self._ser is not None:
            self._ser.close()
            self._ser = None

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate: int) -> None:
        self._baudrate = new_baudrate

        if self._ser is not None and self._ser.is_open:
            self._ser.baudrate = new_baudrate


class ExploreSerialLink(SerialLink):
    _SERIAL_READ_PACKET_SIZE = 65536
    _SERIAL_PACKET_TIMEOUT = 0.01
    _SERIAL_WRITE_TIMEOUT = 1.0

    def __init__(self, port: str, flowcontrol: bool = True) -> None:
        super().__init__(port, flowcontrol)
        self._buf = bytearray()

    def _update_timeout(self) -> None:
        pass

    def connect(self) -> None:
        self._ser = serial.Serial(
            timeout=self._SERIAL_PACKET_TIMEOUT,
            write_timeout=self._SERIAL_WRITE_TIMEOUT,
            exclusive=True,
        )
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._ser.open()
        self._buf = bytearray()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

        self.send_break()

    def recv(self, num_bytes: int) -> bytes:
        assert self._ser is not None
        t0 = time()
        while len(self._buf) < num_bytes:
            if time() - t0 > self._timeout:
                msg = "recv timeout"
                raise LinkError(msg)

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        assert self._ser is not None
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                msg = "recv timeout"
                raise LinkError(msg)

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data


class SerialProcessLink(BaseSerialLink):
    def __init__(self, port: Optional[int] = None) -> None:
        super().__init__()
        self._port = port
        self._process: Optional[mp.Process] = None

    def _update_timeout(self) -> None:
        pass

    def connect(self) -> None:
        self._recv_queue: "mp.Queue[bytes]" = mp.Queue()
        self._send_queue: "mp.Queue[Union[Tuple[str, int], bytes]]" = mp.Queue()
        self._flow_event = mp.Event()
        self._error_event = mp.Event()

        args = (
            self._port,
            self._baudrate,
            self._recv_queue,
            self._send_queue,
            self._flow_event,
            self._error_event,
        )

        self._process = mp.Process(
            target=serial_process_program,
            args=args,
            daemon=True,
        )

        if self._process is None:
            msg = "failed to create subprocess"
            raise LinkError(msg)

        self._process.start()

        flow_event_was_set = self._flow_event.wait(self._timeout)

        if flow_event_was_set:
            log.debug("connect - flow event was set")
        else:
            log.debug("connect - flow event was not set (timeout)")
            self.disconnect()
            msg = "failed to connect, timeout"
            raise LinkError(msg)

        if self._error_event.is_set():
            log.debug("connect - error event was set")
            self.disconnect()
            msg = "failed to connect, see traceback from serial process"
            raise LinkError(msg)

        log.debug("connect - successful")

        self._buf = bytearray()

    def recv(self, num_bytes: int) -> bytes:
        self.__empty_queue_into_buf()

        t0 = time()
        while len(self._buf) < num_bytes:
            self.__get_into_buf()

            if time() - t0 > self._timeout:
                msg = "recv timeout"
                raise LinkError(msg)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        self.__empty_queue_into_buf()

        n = len(bs)
        si = 0
        t0 = time()
        while True:
            buf_size = len(self._buf)
            if buf_size >= n:
                try:
                    i = self._buf.index(bs, si)
                except ValueError:
                    pass
                else:
                    break

                si = buf_size - n + 1

            if time() - t0 > self._timeout:
                msg = "recv timeout"
                raise LinkError(msg)

            self.__get_into_buf()

        data = self._buf[: i + n]
        self._buf = self._buf[i + n :]
        return data

    def send(self, data: bytes) -> None:
        self._send_queue.put(data)

    def disconnect(self) -> None:
        if self._process is not None:
            if self._process.exitcode is None:
                self._flow_event.clear()
                self._process.join(1)

            if self._process.exitcode is None:
                log.info("forcing disconnect...")
                self._process.terminate()
                self._process.join(1)
                log.info("forced disconnect successful")

            if self._process.exitcode is None:
                msg = "failed to disconnect"
                raise LinkError(msg)

    def __empty_queue_into_buf(self) -> None:
        while True:
            try:
                data = self._recv_queue.get_nowait()
            except queue.Empty:
                break

            self._buf.extend(data)

    def __get_into_buf(self) -> None:
        try:
            data = self._recv_queue.get(timeout=self._timeout)
        except queue.Empty:
            msg = "recv timeout"
            raise LinkError(msg)
        except InterruptedError:
            return  # fixes interrupt issue on Windows

        self._buf.extend(data)

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate: int) -> None:
        self._baudrate = new_baudrate

        if self._process is not None and self._process.exitcode is None:
            log.debug("Changing baudrate to {}".format(new_baudrate))
            self._send_queue.put(("baudrate", new_baudrate))


def serial_process_program(
    port: int,
    baud: int,
    recv_q: "mp.Queue[bytes]",
    send_q: "mp.Queue[Union[Tuple[str, int], bytes]]",
    flow_event: th.Event,
    error_event: th.Event,
) -> None:
    log.debug("serial communication process started")
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    try:
        _serial_process_program(port, baud, recv_q, send_q, flow_event, error_event)
    except Exception:
        error_event.set()
        flow_event.set()
        sleep(0.1)  # give the main process some time to print log messages
        print("Exception raised in serial process:\n")
        traceback.print_exc()
        print("\n\n")

    recv_q.close()
    send_q.close()


def _serial_process_program(
    port: int,
    baud: int,
    recv_q: "mp.Queue[bytes]",
    send_q: "mp.Queue[Union[Tuple[str, int], bytes]]",
    flow_event: th.Event,
    error_event: th.Event,
) -> None:
    ser = serial.Serial(port=port, baudrate=baud, timeout=0, exclusive=True)
    flow_event.set()
    while flow_event.is_set():
        received = bytearray()
        while True:
            data = bytearray(ser.read(4096))
            if len(data) == 0:
                break
            received.extend(data)
        if len(received) > 0:
            recv_q.put(received)

        sent = False
        while True:
            try:
                x = send_q.get_nowait()
            except queue.Empty:
                break

            if isinstance(x, tuple):
                _, val = x  # assume its a baudrate change
                ser.baudrate = val
            else:
                ser.write(x)
                sent = True

        if not sent:
            sleep(0.0025)

    ser.close()



# Copyright (c) Acconeer AB, 2023
# All rights reserved


from ._detector import DetectorConfig


def get_default_config() -> DetectorConfig:
    return DetectorConfig()


def get_traffic_config() -> DetectorConfig:
    return DetectorConfig(
        start_point=2400,
        num_bins=200,
        max_speed=50.0,
        threshold=10.0,
    )



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

import json
import warnings

from PySide6.QtCore import QThread

from acconeer.exptool.a111 import _modes
from acconeer.exptool.a111.recording import Recorder


warnings.filterwarnings("ignore")


class DataProcessing:
    hist_len = 500

    def prepare_processing(self, parent, params, session_info):
        self.parent = parent
        self.gui_handle = self.parent.parent
        self.sensor_config = params["sensor_config"]
        self.processing_config = params["service_params"]
        self.multi_sensor = params["multi_sensor"]
        self.calibration = params.get("calibration")

        hist_len = params["sweep_buffer"]
        if hist_len is not None and hist_len < 1:
            hist_len = None

        if isinstance(self.processing_config, dict):  # Legacy
            self.processing_config["processing_handle"] = self
            record_processing_config = None
        else:
            record_processing_config = self.processing_config

        self.session_info = session_info

        self.recorder = Recorder(
            sensor_config=self.sensor_config,
            session_info=session_info,
            module_key=params["module_info"].key,
            processing_config=record_processing_config,
            rss_version=params.get("rss_version", None),
            max_len=hist_len,
        )

        self.init_vars()

    def abort_processing(self):
        self.abort = True

    def init_vars(self):
        self.abort = False
        self.first_run = True

    def process(self, unsqueezed_data, info, do_record=True):
        if self.multi_sensor:
            in_data = unsqueezed_data
            in_info = info
        else:
            assert unsqueezed_data.shape[0] == 1
            in_data = unsqueezed_data[0]
            in_info = info[0]

        if self.first_run:
            ext = self.gui_handle.external
            processing_config = self.processing_config
            try:
                self.external = ext(
                    self.sensor_config,
                    processing_config,
                    self.session_info,
                    self.calibration,
                )
            except TypeError as te:
                msg = f"Could not instantiate {ext.__name__}"
                raise TypeError(msg) from te
            self.first_run = False

        out_data = self.external.process(in_data, in_info)
        if out_data is not None:
            self.draw_canvas(out_data)
            if isinstance(out_data, dict) and out_data.get("send_process_data") is not None:
                self.parent._emit("process_data", "", out_data["send_process_data"])

        if do_record:
            self.recorder.sample(info, unsqueezed_data)

        return out_data, self.recorder.record

    def process_saved_data(self, record, parent):
        self.parent = parent
        self.init_vars()

        sensor_config_dict = json.loads(record.sensor_config_dump)
        rate = sensor_config_dict.get("update_rate", None)
        if rate is None and record.mode == _modes.Mode.SPARSE:
            sweep_rate = sensor_config_dict.get("sweep_rate", None)
            if sweep_rate is None:
                sweep_rate = record.session_info.get("sweep_rate", None)
            if sweep_rate is not None:
                rate = sweep_rate / sensor_config_dict.get("sweeps_per_frame", 16)

        if rate is not None:
            sleep_time_ms = int(round(1000 / rate))
        else:
            sleep_time_ms = 3

        selected_sensors = self.sensor_config.sensor
        stored_sensors = sensor_config_dict["sensor"]

        sensor_list = []
        matching_sensors = []
        for sensor_idx in stored_sensors:
            if sensor_idx in selected_sensors:
                sensor_list.append(stored_sensors.index(sensor_idx))
                matching_sensors.append(sensor_idx)

        if not len(sensor_list) or len(matching_sensors) != len(selected_sensors):
            error = "Data is not available for all selected sensors {}.\n".format(selected_sensors)
            error += "I have data for sensors {} ".format(stored_sensors)
            self.parent._emit("set_sensors", "", stored_sensors)
            self.parent._emit("sensor_selection_error", error)
            return

        self.sensor_config.sensor = matching_sensors

        for subinfo, subdata in record:
            if self.abort:
                break

            QThread.msleep(sleep_time_ms)

            subdata = subdata[sensor_list]
            self.process(subdata, subinfo, do_record=False)
            self.parent._emit("sweep_info", "", subinfo)

    def draw_canvas(self, plot_data):
        self.parent._emit("update_external_plots", "", plot_data)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

from typing import Any, Optional, Tuple, Union

import numpy as np

import acconeer.exptool as et

from .client_factory import ClientFactory
from .common import Link, LinkArg, ProtocolArg
from .json.client import SocketClient
from .mock.client import MockClient
from .reg.client import SPIClient, UARTClient


class Client:
    def __init__(
        self,
        *,
        protocol: Optional[ProtocolArg] = None,
        link: Optional[LinkArg] = None,
        mock: bool = False,
        **kwargs: Any,
    ):
        """
        :param protocol:
            What protocol the client is supposed to use. Can be any :class:`.Protocol`
            member or their ``str``-counterparts.
            ``protocol=None`` (default) will try to auto-detect.

        :param link:
            What link the client is supposed to use. Can be any :class:`.Link`
            member or their ``str```-counterparts.
            ``link=None`` (default) will try to auto-detect.

        :param mock:
            Whether this Client should be a simulated client.

        :param kwargs:
            These are the supported kwargs:

            | **host:** str
            |   IP-address of e.g. the RBPi you want to connect to.

            | **serial_port:** str
            |   The serial port name. E.g. ``COMx`` on Windows and ``/dev/ttyUSBx`` on Linux.

            | **override_baudrate:** int
            |   Uses the passed baudrate instead of the default.

        :raises: ValueError if a ``Client`` could not be created from the arguments.
        """
        if mock:
            self.subclient = MockClient()
        else:
            self.subclient = ClientFactory.from_kwargs(protocol=protocol, link=link, **kwargs)

    def get_link_type(self) -> Union[Link, str]:
        if isinstance(self.subclient, MockClient):
            return "mock"
        elif isinstance(self.subclient, UARTClient):
            return Link.UART
        elif isinstance(self.subclient, SPIClient):
            return Link.SPI
        elif isinstance(self.subclient, SocketClient):
            return Link.SOCKET
        else:
            msg = f"Unknown subclient type: {type(self.subclient)}"
            raise ValueError(msg)

    def connect(self) -> dict:
        """Initiates a connection with the device.

        :return: A dict containing information about the device, including SDK version
        """
        return self.subclient.connect()

    def setup_session(
        self, config: et.a111._configs.BaseSessionConfig, check_config: bool = True
    ) -> dict:
        """
        Sets up a session with the given config.
        Will call ``connect()`` if not already connected.

        :param config: The configuration to use when setting up the session
        :param check_config: If `True` the configuration is checked for errors,
                            defaults to `True`

        :return: A dict with metadata for the configured session
        """
        return self.subclient.setup_session(config, check_config)

    def start_session(
        self,
        config: Optional[et.a111._configs.BaseSessionConfig] = None,
        check_config: bool = True,
    ) -> Optional[dict]:
        """
        Starts the session if previously set up with ``setup_session()``.
        If `config` is provided, ``setup_session()`` will be called.

        :param config: The configuration to use when setting up the session, defaults to `None`
        :param check_config: If `True` the configuration is checked for errors,
                            defaults to `True`

        :return: If `config` is provided, returns a dict with metadata for the configured session.
                Otherwise, returns `None`
        """
        return self.subclient.start_session(config, check_config)

    def get_next(self) -> Tuple[Union[list, dict], np.ndarray]:
        """
        Retrieves the next result. Will block until the result is received.

        :return: A tuple with the result info and data.
                The data shape and type differs between services.

                | **Power Bins:**
                | Shape: (number of sensors, bin count)
                | Type: float64

                | **Envelope:**
                | Shape: (number of sensors, data length)
                | Type: float64

                | **IQ:**
                | Shape: (number of sensors, data length)
                | Type: complex128

                | **Sparse:**
                | Shape: (number of sensors, number of sweeps, number of depths)
                | Type: float64

                `Number of sensors`, `bin count` and `number of sweeps` can be explicitly set.
                `Data length` and `number of dephts` depend on multiple configuration settings.

                The client takes a parameter ``squeeze``, if set to `True` the first
                dimension (`number of sensors`) is removed when using a single sensor.
                As default ``squeeze`` is  `True`.
        """
        return self.subclient.get_next()

    def stop_session(self):
        """
        Stops the session. All buffered/waiting data is thrown away.
        This function will block until the server has confirmed that the session has ended.
        """
        self.subclient.stop_session()

    def disconnect(self):
        """
        Disconnects the client. ``disconnect()`` will call ``stop_session()``
        if a session is started.
        """
        self.subclient.disconnect()

    @property
    def supported_modes(self):
        return self.subclient.supported_modes

    @property
    def squeeze(self):
        return self.subclient.squeeze

    @squeeze.setter
    def squeeze(self, squeeze):
        self.subclient.squeeze = squeeze

    @property
    def description(self) -> str:
        return self.subclient.description



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import typing as t

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QWidget

from acconeer.exptool._core.entities.validation_result import ValidationResult


_DataT = t.TypeVar("_DataT")


class DataEditor(t.Generic[_DataT], QWidget):
    """A DataEditor is a widget that enables a user to edit some data.
    The updated data can be used programmatically and is propagated via
    the `sig_update` Signal or retrieved via `get_data`.
    """

    sig_update = Signal(object)

    def set_data(self, data: _DataT) -> None:
        """Update the data that is displayed"""
        raise NotImplementedError

    def get_data(self) -> _DataT:
        """Gets the data stored in the widget"""
        raise NotImplementedError

    def setEnabled(self, enabled: bool) -> None:
        super().setEnabled(enabled and self.get_data() is not None)

    def handle_validation_results(
        self, validation_results: list[ValidationResult]
    ) -> list[ValidationResult]:
        """Handles validation results by displaying them in the editor in some way.

        Returns ValidationResults that could not be handled.
        """
        raise NotImplementedError

    @property
    def is_ready(self) -> bool:
        """Returns true if there are any non-validation errors present."""
        raise NotImplementedError



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved
from __future__ import annotations

import typing as t

from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import processor_main

from . import Processor, ProcessorConfig, get_sensor_config
from ._plugin import PlotPlugin


def _processor_getter(
    session_config: a121.SessionConfig,
    processor_config: ProcessorConfig,
    metadata: t.Union[a121.Metadata, t.List[t.Dict[int, a121.Metadata]]],
) -> Processor:
    if isinstance(metadata, list):
        msg = "Metadata is unexpectedly extended"
        raise RuntimeError(msg)

    return Processor(
        sensor_config=session_config.sensor_config,
        metadata=metadata,
        processor_config=processor_config,
    )


def _session_config_getter(sensor_id: int) -> a121.SessionConfig:
    return a121.SessionConfig({sensor_id: get_sensor_config()})


processor_main(
    processor_getter=_processor_getter,
    plot_plugin=PlotPlugin,
    session_config_getter=_session_config_getter,
    processor_config_getter=ProcessorConfig,
)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
"""
This module provides a simple default value provider,
making it quick to change default values module-wide.
"""

from . import algo, lookup


DEFAULT_ALGO = algo.SparseIq()
DEFAULT_SENSOR = lookup.Sensor.default()
DEFAULT_MODULE = lookup.Module.xm125()



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QHBoxLayout, QWidget

from . import pidgets


class TwoSensorIdsEditor(QWidget):
    sig_update = Signal(object)

    def __init__(self, name_label_texts: list[str]):
        super().__init__()

        self._sensor_id_pidget_1 = pidgets.SensorIdPidgetFactory(
            name_label_text=name_label_texts[0], items=[]
        ).create(self)
        self._sensor_id_pidget_2 = pidgets.SensorIdPidgetFactory(
            name_label_text=name_label_texts[1], items=[]
        ).create(self)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(self._sensor_id_pidget_1)
        layout.addWidget(self._sensor_id_pidget_2)

        self._sensor_id_pidget_1.sig_update.connect(lambda: self.sig_update.emit(self.sensor_ids))
        self._sensor_id_pidget_2.sig_update.connect(lambda: self.sig_update.emit(self.sensor_ids))

        self.setLayout(layout)

    @property
    def sensor_ids(self) -> list[int]:
        return [
            self._sensor_id_pidget_1.get_data(),
            self._sensor_id_pidget_2.get_data(),
        ]

    def set_data(self, sensor_ids: list[int]) -> None:
        assert len(sensor_ids) == 2
        self._sensor_id_pidget_1.set_data(sensor_ids[0])
        self._sensor_id_pidget_2.set_data(sensor_ids[1])

    def set_selectable_sensors(self, sensor_list: list[int]) -> None:
        self._sensor_id_pidget_1.set_selectable_sensors(sensor_list)
        self._sensor_id_pidget_2.set_selectable_sensors(sensor_list)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import logging
from typing import List, Optional, Union

import attr
import yaml

import acconeer.exptool as et
from acconeer.exptool.a111 import algo


log = logging.getLogger(__name__)


def list_flattener(list_or_num: Union[list, float]):
    if isinstance(list_or_num, list):
        if len(list_or_num) != 1:
            msg = f"argument {list_or_num} needs to be a length 1 list or a number."
            raise ValueError(msg)
        return list_or_num[0]
    return list_or_num


has_float_elements = attr.validators.deep_iterable(attr.validators.instance_of(float), None)


@attr.frozen
class ObstacleDetectionCalibration(algo.Calibration):
    """
    Calibration for obstacle detector.

    Supports to be loaded from ".yaml" files.
    """

    static_pwl_dist: List[float] = attr.ib(validator=has_float_elements)
    static_pwl_amp: List[float] = attr.ib(validator=has_float_elements)
    moving_pwl_dist: List[float] = attr.ib(validator=has_float_elements)
    moving_pwl_amp: List[float] = attr.ib(validator=has_float_elements)
    static_adjacent_factor: Union[float, List[float]] = attr.ib(
        validator=attr.validators.instance_of(float),
        converter=attr.converters.pipe(list_flattener, float),
    )
    moving_max: Union[float, List[float]] = attr.ib(
        validator=attr.validators.instance_of(float),
        converter=attr.converters.pipe(list_flattener, float),
    )

    @static_pwl_dist.validator
    @static_pwl_amp.validator
    def check_static_list(self, attribute, list_o_floats):
        if len(list_o_floats) != 5:
            msg = f"{attribute.name} is supposed have length = 5 "
            raise ValueError(msg)

    @moving_pwl_dist.validator
    @moving_pwl_amp.validator
    def check_moving_list(self, attribute, list_o_floats):
        if attribute.name.startswith("moving") and len(list_o_floats) != 2:
            msg = f"{attribute.name} is supposed have length = 2 "
            raise ValueError(msg)

    def save(self, path: algo._Path):
        self.validate_path(path)

        with open(path, "w") as stream:
            yaml.dump(
                dict(
                    static_pwl_dist=self.static_pwl_dist,
                    static_pwl_amp=self.static_pwl_amp,
                    moving_pwl_dist=self.moving_pwl_dist,
                    moving_pwl_amp=self.moving_pwl_amp,
                    static_adjacent_factor=self.static_adjacent_factor,
                    moving_max=self.moving_max,
                ),
                stream=stream,
                default_flow_style=False,
            )

    @classmethod
    def load(cls, path: algo._Path) -> algo.Calibration:
        cls.validate_path(path)

        with open(path, "r") as stream:
            calibration_dict = yaml.safe_load(stream)

        return cls(**calibration_dict)

    @classmethod
    def file_extensions(cls) -> algo._AcceptedFileExtensions:
        return [("yaml", "YAML files (*.yaml)")]


class ObstacleDetectionCalibrationConfiguration(et.configbase.Config):
    VERSION = 1

    static_adjacent_factor = et.configbase.FloatParameter(
        label="Static adjacent factor",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=0,
    )

    moving_max = et.configbase.FloatParameter(
        label="Moving max",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=10,
    )

    static_pwl_dist_1 = et.configbase.FloatParameter(
        label="Static PWL distance 1",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=20,
    )
    static_pwl_dist_2 = et.configbase.FloatParameter(
        label="Static PWL distance 2",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=21,
    )
    static_pwl_dist_3 = et.configbase.FloatParameter(
        label="Static PWL distance 3",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=22,
    )
    static_pwl_dist_4 = et.configbase.FloatParameter(
        label="Static PWL distance 4",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=23,
    )
    static_pwl_dist_5 = et.configbase.FloatParameter(
        label="Static PWL distance 5",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=24,
    )

    static_pwl_amp_1 = et.configbase.FloatParameter(
        label="Static PWL amplitude 1",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=30,
    )
    static_pwl_amp_2 = et.configbase.FloatParameter(
        label="Static PWL amplitude 2",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=31,
    )
    static_pwl_amp_3 = et.configbase.FloatParameter(
        label="Static PWL amplitude 3",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=32,
    )
    static_pwl_amp_4 = et.configbase.FloatParameter(
        label="Static PWL amplitude 4",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=33,
    )
    static_pwl_amp_5 = et.configbase.FloatParameter(
        label="Static PWL amplitude 5",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=34,
    )

    moving_pwl_dist_1 = et.configbase.FloatParameter(
        label="Moving PWL distance 1",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=40,
    )
    moving_pwl_dist_2 = et.configbase.FloatParameter(
        label="Moving PWL distance 2",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=41,
    )

    moving_pwl_amp_1 = et.configbase.FloatParameter(
        label="Moving PWL amplitude 1",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=50,
    )
    moving_pwl_amp_2 = et.configbase.FloatParameter(
        label="Moving PWL amplitude 2",
        default_value=0.0,
        limits=(0, None),
        decimals=6,
        order=51,
    )


class ObstacleDetectionCalibrationMapper(algo.CalibrationMapper):
    @classmethod
    def get_updated_calibration_from_configuration(
        cls,
        configuration: ObstacleDetectionCalibrationConfiguration,
        calibration: Optional[ObstacleDetectionCalibration],
    ) -> ObstacleDetectionCalibration:
        """
        Creates a new calibration given a configuration and (optionally) a calibration.

        The calibration is optional in order to allow for unsymmetric mapping between
        CalibrationConfiguration and Calibration, where fields not covered by the
        CalibrationConfiguration are copies from the Calibration. (see CalibrationMapper)
        """

        return ObstacleDetectionCalibration(
            static_pwl_dist=[
                configuration.static_pwl_dist_1,
                configuration.static_pwl_dist_2,
                configuration.static_pwl_dist_3,
                configuration.static_pwl_dist_4,
                configuration.static_pwl_dist_5,
            ],
            static_pwl_amp=[
                configuration.static_pwl_amp_1,
                configuration.static_pwl_amp_2,
                configuration.static_pwl_amp_3,
                configuration.static_pwl_amp_4,
                configuration.static_pwl_amp_5,
            ],
            moving_pwl_dist=[
                configuration.moving_pwl_dist_1,
                configuration.moving_pwl_dist_2,
            ],
            moving_pwl_amp=[
                configuration.moving_pwl_amp_1,
                configuration.moving_pwl_amp_2,
            ],
            static_adjacent_factor=configuration.static_adjacent_factor,
            moving_max=configuration.moving_max,
        )

    @classmethod
    def update_config_from_calibration(
        cls,
        configuration: ObstacleDetectionCalibrationConfiguration,
        calibration: ObstacleDetectionCalibration,
    ) -> None:
        """
        Updates fields of CalibrationConfiguration given a ObstacleDetectionCalibration
        """
        s_dist = calibration.static_pwl_dist
        configuration.static_pwl_dist_1 = s_dist[0]
        configuration.static_pwl_dist_2 = s_dist[1]
        configuration.static_pwl_dist_3 = s_dist[2]
        configuration.static_pwl_dist_4 = s_dist[3]
        configuration.static_pwl_dist_5 = s_dist[4]

        s_amp = calibration.static_pwl_amp
        configuration.static_pwl_amp_1 = s_amp[0]
        configuration.static_pwl_amp_2 = s_amp[1]
        configuration.static_pwl_amp_3 = s_amp[2]
        configuration.static_pwl_amp_4 = s_amp[3]
        configuration.static_pwl_amp_5 = s_amp[4]

        m_dist = calibration.moving_pwl_dist
        configuration.moving_pwl_dist_1 = m_dist[0]
        configuration.moving_pwl_dist_2 = m_dist[1]

        m_amp = calibration.moving_pwl_amp
        configuration.moving_pwl_amp_1 = m_amp[0]
        configuration.moving_pwl_amp_2 = m_amp[1]

        configuration.moving_max = calibration.moving_max
        configuration.static_adjacent_factor = calibration.static_adjacent_factor
